import json

import pytest

from jobrunner import config
from jobrunner.cli import manifest_migration
from jobrunner.lib import database
from jobrunner.manage_jobs import MANIFEST_FILE, METADATA_DIR
from jobrunner.models import Job


def test_triggers_migration(tmp_work_dir):
    write_manifest(
        {
            "workspace": "the-workspace",
            "actions": {"the-action": {"job_id": "job-id-from-manifest"}},
            "files": {},
        }
    )

    manifest_migration.main()

    assert database.exists_where(Job, id="job-id-from-manifest")


def test_passes_on_batch_size(tmp_work_dir):
    write_manifest(
        {
            "workspace": "the-workspace",
            "actions": {f"action-{i}": {"job_id": f"job-{i}"} for i in range(100)},
            "files": {},
        }
    )

    manifest_migration.main(["--batch-size", "5"])

    assert len(database.find_where(Job, workspace="the-workspace")) == 5


def test_defaults_batch_size_to_one(tmp_work_dir):
    write_manifest(
        {
            "workspace": "the-workspace",
            "actions": {f"action-{i}": {"job_id": f"job-{i}"} for i in range(5)},
            "files": {},
        }
    )

    manifest_migration.main()

    assert len(database.find_where(Job, workspace="the-workspace")) == 1


def test_passes_on_dry_run(tmp_work_dir):
    write_manifest(
        {
            "workspace": "the-workspace",
            "actions": {"the-action": {"job_id": "job-id-from-manifest"}},
            "files": {},
        }
    )

    manifest_migration.main(["--dry-run"])

    assert not database.exists_where(Job, id="job-id-from-manifest")


def test_ignores_errors_for_dry_run(tmp_work_dir):
    # This manifest is invalid because it doesn't contain an "actions" element.
    write_manifest(
        {
            "workspace": "the-workspace",
            "files": {},
        }
    )

    # Future-proof the test against implementation changes. If this fails it's because the manifest above is no longer
    # invalid, so we need to reformulate the test.
    with pytest.raises(KeyError):
        manifest_migration.main()

    # No exception should be raised with --dry-run.
    manifest_migration.main(["--dry-run"])


def write_manifest(manifest):
    manifest_file = (
        config.HIGH_PRIVACY_WORKSPACES_DIR
        / "the-workspace"
        / METADATA_DIR
        / MANIFEST_FILE
    )
    manifest_file.parent.mkdir(parents=True)
    manifest_file.write_text(json.dumps(manifest))



"""
This module contains the logic for starting jobs in Docker containers and
dealing with them when they are finished.

It's important that the `start_job` and `finalise_job` functions are
idempotent. This means that the job-runner can be killed at any point and will
still end up in a consistent state when it's restarted.
"""
import datetime
import json
import logging
import os.path
import shlex
import shutil
import tempfile
import time
from pathlib import Path

from jobrunner import config
from jobrunner.lib import atomic_writer, docker
from jobrunner.lib.database import find_one
from jobrunner.lib.git import checkout_commit
from jobrunner.lib.path_utils import list_dir_with_ignore_patterns
from jobrunner.lib.string_utils import tabulate
from jobrunner.lib.subprocess_utils import subprocess_run
from jobrunner.models import SavedJobRequest, State, StatusCode
from jobrunner.project import (
    get_all_output_patterns_from_project_file,
    requires_db_access,
)
from jobrunner.queries import calculate_workspace_state


log = logging.getLogger(__name__)

# Directory inside working directory where manifest and logs are created
METADATA_DIR = "metadata"

# Records details of which action created each file
MANIFEST_FILE = "manifest.json"

# Keys of fields to log in manifest.json and log file
KEYS_TO_LOG = [
    "state",
    "commit",
    "docker_image_id",
    "action_repo_url",
    "action_commit",
    "job_id",
    "run_by_user",
    "created_at",
    "completed_at",
    "exit_code",
]


# This is a Docker label applied in addition to the default label which
# `docker.py` applies to all containers and volumes it creates. It allows us to
# easily identify just the containers actually used for running jobs, which is
# helpful for building tooling for inspecting live processes.
JOB_LABEL = "jobrunner-job"

# This is part of a hack we use to track which files in a volume are newly
# created
TIMESTAMP_REFERENCE_FILE = ".opensafely-timestamp"


class JobError(Exception):
    pass


class ActionNotRunError(JobError):
    pass


class ActionFailedError(JobError):
    pass


class MissingOutputError(JobError):
    pass


class BrokenContainerError(JobError):
    pass


def start_job(job):
    """Start the given job.

    Args:
        job: An instance of Job.
    """
    # If we already created the job but were killed before we updated the state
    # then there's nothing further to do
    if docker.container_exists(container_name(job)):
        log.info("Container already created, nothing to do")
        return
    try:
        volume = create_and_populate_volume(job)
    except docker.DockerDiskSpaceError as e:
        log.exception(str(e))
        raise JobError("Out of disk space, please try again later")
    action_args = shlex.split(job.run_command)
    allow_network_access = False
    env = {"OPENSAFELY_BACKEND": config.BACKEND}
    # Check `is True` so we fail closed if we ever get anything else
    if requires_db_access(action_args) is True:
        if not config.USING_DUMMY_DATA_BACKEND:
            allow_network_access = True
            env["DATABASE_URL"] = config.DATABASE_URLS[job.database_name]
            if config.TEMP_DATABASE_NAME:
                env["TEMP_DATABASE_NAME"] = config.TEMP_DATABASE_NAME
            if config.PRESTO_TLS_KEY and config.PRESTO_TLS_CERT:
                env["PRESTO_TLS_CERT"] = config.PRESTO_TLS_CERT
                env["PRESTO_TLS_KEY"] = config.PRESTO_TLS_KEY
            if config.EMIS_ORGANISATION_HASH:
                env["EMIS_ORGANISATION_HASH"] = config.EMIS_ORGANISATION_HASH
    # Prepend registry name
    image = action_args[0]
    full_image = f"{config.DOCKER_REGISTRY}/{image}"
    if image.startswith("stata-mp"):
        env["STATA_LICENSE"] = str(config.STATA_LICENSE)
    # Check the image exists locally and error if not. Newer versions of
    # docker-cli support `--pull=never` as an argument to `docker run` which
    # would make this simpler, but it looks like it will be a while before this
    # makes it to Docker for Windows:
    # https://github.com/docker/cli/pull/1498
    if not docker.image_exists_locally(full_image):
        log.info(f"Image not found, may need to run: docker pull {full_image}")
        raise JobError(f"Docker image {image} is not currently available")
    # Start the container
    docker.run(
        container_name(job),
        [full_image] + action_args[1:],
        volume=(volume, "/workspace"),
        env=env,
        allow_network_access=allow_network_access,
        label=JOB_LABEL,
    )
    log.info("Started")
    log.info(f"View live logs using: docker logs -f {container_name(job)}")


def create_and_populate_volume(job):
    workspace_dir = get_high_privacy_workspace(job.workspace)
    input_files = {}
    for action in job.requires_outputs_from:
        for filename in list_outputs_from_action(job.workspace, action):
            input_files[filename] = action

    volume = volume_name(job)
    docker.create_volume(volume)

    # `docker cp` can't create parent directories for us so we make sure all
    # these directories get created when we copy in the code
    extra_dirs = set(Path(filename).parent for filename in input_files.keys())

    # Jobs which are running reusable actions pull their code from the reusable
    # action repo, all other jobs pull their code from the study repo
    repo_url = job.action_repo_url or job.repo_url
    commit = job.action_commit or job.commit
    # Both of action commit and repo_url should be set if either are
    assert bool(job.action_commit) == bool(job.action_repo_url)

    if repo_url and commit:
        copy_git_commit_to_volume(volume, repo_url, commit, extra_dirs)
    else:
        # We only encounter jobs without a repo or commit when using the
        # "local_run" command to execute uncommitted local code
        copy_local_workspace_to_volume(volume, workspace_dir, extra_dirs)

    for filename, action in input_files.items():
        log.info(f"Copying input file {action}: {filename}")
        docker.copy_to_volume(volume, workspace_dir / filename, filename)
    # Hack: see `get_unmatched_outputs`. For some reason this requires a
    # non-empty file so copying `os.devnull` didn't work.
    some_non_empty_file = Path(__file__)
    docker.copy_to_volume(volume, some_non_empty_file, TIMESTAMP_REFERENCE_FILE)
    return volume


def copy_git_commit_to_volume(volume, repo_url, commit, extra_dirs):
    log.info(f"Copying in code from {repo_url}@{commit}")
    # git-archive will create a tarball on stdout and docker cp will accept a
    # tarball on stdin, so if we wanted to we could do this all without a
    # temporary directory, but not worth it at this stage
    config.TMP_DIR.mkdir(parents=True, exist_ok=True)
    with tempfile.TemporaryDirectory(dir=config.TMP_DIR) as tmpdir:
        tmpdir = Path(tmpdir)
        checkout_commit(repo_url, commit, tmpdir)
        # Because `docker cp` can't create parent directories automatically, we
        # make sure parent directories exist for all the files we're going to
        # copy in later
        for directory in extra_dirs:
            tmpdir.joinpath(directory).mkdir(parents=True, exist_ok=True)
        try:
            docker.copy_to_volume(volume, tmpdir, ".", timeout=60)
        except docker.DockerTimeoutError:
            # Aborting a `docker cp` into a container at the wrong time can
            # leave the container in a completely broken state where any
            # attempt to interact with or even remove it will just hang, see:
            # https://github.com/docker/for-mac/issues/4491
            #
            # This means we can end up with jobs where any attempt to start
            # them (by copying in code from git) causes the job-runner to
            # completely lock up. To avoid this we use a timeout (60 seconds,
            # which should be more than enough to copy in a few megabytes of
            # code). The exception this triggers will cause the job to fail
            # with an "internal error" message, which will then stop it
            # blocking other jobs. We need a specific exception class here as
            # we need to avoid trying to remove the container, which we would
            # ordinarily do on error, because that operation will also hang :(
            log.exception("Timed out copying code to volume, see issue #154")
            raise BrokenContainerError(
                "There was a (hopefully temporary) internal Docker error, "
                "please try the job again"
            )


def copy_local_workspace_to_volume(volume, workspace_dir, extra_dirs):
    # To mimic a production run, we only want output files to appear in the
    # volume if they were produced by an explicitly listed dependency. So
    # before copying in the code we get a list of all output patterns in the
    # project and ignore any files matching these patterns
    project_file = workspace_dir / "project.yaml"
    ignore_patterns = get_all_output_patterns_from_project_file(project_file)
    ignore_patterns.extend([".git", METADATA_DIR])
    code_files = list_dir_with_ignore_patterns(workspace_dir, ignore_patterns)

    # Because `docker cp` can't create parent directories automatically, we
    # need to make sure empty parent directories exist for all the files we're
    # going to copy in. For now we do this by actually creating a bunch of
    # empty dirs in a temp directory. It should be possible to do this using
    # the `tarfile` module to talk directly to `docker cp` stdin if we care
    # enough.
    directories = set(Path(filename).parent for filename in code_files)
    directories.update(extra_dirs)
    directories.discard(Path("."))
    if directories:
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            for directory in directories:
                tmpdir.joinpath(directory).mkdir(parents=True, exist_ok=True)
            docker.copy_to_volume(volume, tmpdir, ".")

    log.info(f"Copying in code from {workspace_dir}")
    for filename in code_files:
        docker.copy_to_volume(volume, workspace_dir / filename, filename)


def job_still_running(job):
    return docker.container_is_running(container_name(job))


# note: these functions use different naming logic for to isolate the state between the old/new worlds
def container_name(job):
    return f"os-job-{job.id}" if config.EXECUTION_API else f"job-{job.slug}"


def volume_name(job):
    return f"os-volume-{job.id}" if config.EXECUTION_API else f"volume-{job.slug}"


def finalise_job(job):
    """
    This involves checking whether the job finished successfully or not and
    extracting all outputs, logs and metadata
    """
    container_metadata = get_container_metadata(job)
    outputs, unmatched_patterns = find_matching_outputs(job)
    job.outputs = outputs
    job.image_id = container_metadata["Image"]

    # Set the final state of the job
    if container_metadata["State"]["ExitCode"] != 0:
        job.state = State.FAILED
        job.status_message = "Job exited with an error code"
        job.status_code = StatusCode.NONZERO_EXIT
    elif unmatched_patterns:
        job.state = State.FAILED
        job.status_message = "No outputs found matching patterns:\n - {}".format(
            "\n - ".join(unmatched_patterns)
        )
        # If the job fails because an output was missing its very useful to
        # show the user what files were created as often the issue is just a
        # typo
        job.unmatched_outputs = get_unmatched_outputs(job)
    else:
        job.state = State.SUCCEEDED
        job.status_message = "Completed successfully"

    # job_metadata is a big dict capturing everything we know about the state
    # of the job
    job_metadata = get_job_metadata(job, container_metadata)

    # Dump useful info in log directory
    log_dir = get_log_dir(job)
    ensure_overwritable(log_dir / "logs.txt", log_dir / "metadata.json")
    write_log_file(job, job_metadata, log_dir / "logs.txt")
    with open(log_dir / "metadata.json", "w") as f:
        json.dump(job_metadata, f, indent=2)

    # If a job was cancelled we bail out before making any changes to the
    # workspace, but after having written the log and metadata files to the
    # long-term logs directory for debugging purposes.
    if job.cancelled:
        raise JobError("Cancelled by user")

    # Copy logs to workspace
    workspace_dir = get_high_privacy_workspace(job.workspace)
    metadata_log_file = workspace_dir / METADATA_DIR / f"{job.action}.log"
    copy_file(log_dir / "logs.txt", metadata_log_file)
    log.info(f"Logs written to: {metadata_log_file}")

    # Extract outputs to workspace
    ensure_overwritable(*[workspace_dir / f for f in job.output_files])
    volume = volume_name(job)
    for filename in job.output_files:
        log.info(f"Extracting output file: {filename}")
        docker.copy_from_volume(volume, filename, workspace_dir / filename)

    # Delete outputs from previous run of action. It would be simpler to delete
    # all existing outputs and then copy over the new ones, but this way we
    # don't delete anything until after we've copied the new outputs which is
    # safer in case anything goes wrong.
    existing_files = list_outputs_from_action(job.workspace, job.action)
    delete_files(workspace_dir, existing_files, files_to_keep=job.output_files)

    # Copy out logs and medium privacy files
    medium_privacy_dir = get_medium_privacy_workspace(job.workspace)
    if medium_privacy_dir:
        copy_file(
            workspace_dir / METADATA_DIR / f"{job.action}.log",
            medium_privacy_dir / METADATA_DIR / f"{job.action}.log",
        )
        new_files = []
        for filename, privacy_level in job.outputs.items():
            if privacy_level == "moderately_sensitive":
                copy_file(workspace_dir / filename, medium_privacy_dir / filename)
                new_files.append(filename)
        delete_files(medium_privacy_dir, existing_files, files_to_keep=new_files)

        # osrelease needs to be able to read the workspace name and repo URL from somewhere, in order to avoid the
        # person doing the release having to enter all the details. So we write this rump manifest just into the
        # medium privacy workspace. release-hatch is launched with this information already provided, so when osrelease
        # has been removed we can stop doing this.
        #
        # We only really need to write this the first time that an action is run in a workspace, but it's easier to do
        # it here every time than try to detect that.
        write_manifest_file(
            medium_privacy_dir, {"repo": job.repo_url, "workspace": job.workspace}
        )

    return job


def cleanup_job(job):
    if config.CLEAN_UP_DOCKER_OBJECTS:
        log.info("Cleaning up container and volume")
        docker.delete_container(container_name(job))
        docker.delete_volume(volume_name(job))
    else:
        log.info("Leaving container and volume in place for debugging")


def kill_job(job):
    docker.kill(container_name(job))


def get_container_metadata(job):
    metadata = docker.container_inspect(container_name(job), none_if_not_exists=True)
    if not metadata:
        raise JobError("Job container has vanished")
    redact_environment_variables(metadata)
    return metadata


def find_matching_outputs(job):
    """
    Returns a dict mapping output filenames to their privacy level, plus a list
    of any patterns that had no matches at all
    """
    all_patterns = []
    for privacy_level, named_patterns in job.output_spec.items():
        for name, pattern in named_patterns.items():
            all_patterns.append(pattern)
    all_matches = docker.glob_volume_files(volume_name(job), all_patterns)
    unmatched_patterns = []
    outputs = {}
    for privacy_level, named_patterns in job.output_spec.items():
        for name, pattern in named_patterns.items():
            filenames = all_matches[pattern]
            if not filenames:
                unmatched_patterns.append(pattern)
            for filename in filenames:
                outputs[filename] = privacy_level
    return outputs, unmatched_patterns


def get_unmatched_outputs(job):
    """
    Returns all the files created by the job which were *not* matched by any of
    the output patterns.

    This is very useful in debugging because it's easy for users to get their
    output patterns wrong (often just in the wrong directory) and it becomes
    immediately obvious what the problem is if, as well as an error, they can
    see a list of the files the job *did* produce.

    The way we do this is bit hacky, but given that it's only used for
    debugging info and not for Serious Business Purposes, it should be
    sufficient.
    """
    all_outputs = docker.find_newer_files(volume_name(job), TIMESTAMP_REFERENCE_FILE)
    return [filename for filename in all_outputs if filename not in job.outputs]


def get_job_metadata(job, container_metadata):
    """
    Returns a JSON-serializable dict including everything we know about a job
    """
    # This won't exactly match the final `completed_at` time which doesn't get
    # set until the entire job has finished processing, but we want _some_ kind
    # of time to put in the metadata
    job.completed_at = int(time.time())
    job_metadata = job.asdict()
    job_request = find_one(SavedJobRequest, id=job.job_request_id)
    # The original job_request, exactly as received from the job-server
    job_metadata["job_request"] = job_request.original
    job_metadata["job_id"] = job_metadata["id"]
    job_metadata["run_by_user"] = job_metadata["job_request"].get("created_by")
    job_metadata["docker_image_id"] = container_metadata["Image"]
    job_metadata["exit_code"] = container_metadata["State"]["ExitCode"]
    job_metadata["container_metadata"] = container_metadata
    return job_metadata


def write_log_file(job, job_metadata, filename):
    """
    This dumps the (timestamped) Docker logs for a job to disk, followed by
    some useful metadata about the job and its outputs
    """
    filename.parent.mkdir(parents=True, exist_ok=True)
    docker.write_logs_to_file(container_name(job), filename)
    outputs = sorted(job_metadata["outputs"].items())
    with open(filename, "a") as f:
        f.write("\n\n")
        for key in KEYS_TO_LOG:
            if not job_metadata[key]:
                continue
            f.write(f"{key}: {job_metadata[key]}\n")
        f.write(f"\n{job_metadata['status_message']}\n")
        if job.unmatched_outputs:
            f.write("\nDid you mean to match one of these files instead?\n - ")
            f.write("\n - ".join(job.unmatched_outputs))
            f.write("\n")
        f.write("\noutputs:\n")
        f.write(tabulate(outputs, separator="  - ", indent=2, empty="(no outputs)"))


# Environment variables whose values do not need to be hidden from the debug
# logs
SAFE_ENVIRONMENT_VARIABLES = set(
    """
    PATH PYTHON_VERSION DEBIAN_FRONTEND DEBCONF_NONINTERACTIVE_SEEN
    UBUNTU_VERSION PYENV_SHELL PYENV_VERSION PYTHONUNBUFFERED
    OPENSAFELY_BACKEND TZ TEMP_DATABASE_NAME PYTHONPATH container LANG LC_ALL
    """.split()
)


def redact_environment_variables(container_metadata):
    """
    Redact the values of any environment variables in the container which
    aren't on the explicit safelist
    """
    env_vars = [line.split("=", 1) for line in container_metadata["Config"]["Env"]]
    redacted_vars = [
        f"{key}=xxxx-REDACTED-xxxx"
        if key not in SAFE_ENVIRONMENT_VARIABLES
        else f"{key}={value}"
        for (key, value) in env_vars
    ]
    container_metadata["Config"]["Env"] = redacted_vars


def get_log_dir(job):
    # Split log directory up by month to make things slightly more manageable
    month_dir = datetime.date.today().strftime("%Y-%m")
    return config.JOB_LOG_DIR / month_dir / container_name(job)


def copy_file(source, dest):
    dest.parent.mkdir(parents=True, exist_ok=True)
    ensure_overwritable(dest)

    with atomic_writer(dest) as tmp:
        shutil.copy(source, tmp)


def delete_files(directory, filenames, files_to_keep=()):
    ensure_overwritable(*[directory.joinpath(f) for f in filenames])
    # We implement the "files to keep" logic using inodes rather than names so
    # we can safely handle case-insensitive filesystems
    inodes_to_keep = set()
    for filename in files_to_keep:
        try:
            stat = directory.joinpath(filename).stat()
            inodes_to_keep.add((stat.st_dev, stat.st_ino))
        except FileNotFoundError:
            pass
    for filename in filenames:
        path = directory / filename
        try:
            stat = path.stat()
        except FileNotFoundError:
            continue
        inode = (stat.st_dev, stat.st_ino)
        if inode not in inodes_to_keep:
            path.unlink()


def list_outputs_from_action(workspace, action):
    for job in calculate_workspace_state(workspace):
        if job.action == action:
            return job.output_files

    # The action has never been run before
    return []


def write_manifest_file(workspace_dir, manifest):
    manifest_file = workspace_dir / METADATA_DIR / MANIFEST_FILE
    manifest_file_tmp = manifest_file.with_suffix(".tmp")
    ensure_overwritable(manifest_file, manifest_file_tmp)
    manifest_file_tmp.write_text(json.dumps(manifest, indent=2))
    manifest_file_tmp.replace(manifest_file)


def get_high_privacy_workspace(workspace):
    return config.HIGH_PRIVACY_WORKSPACES_DIR / workspace


def get_medium_privacy_workspace(workspace):
    if config.MEDIUM_PRIVACY_WORKSPACES_DIR:
        return config.MEDIUM_PRIVACY_WORKSPACES_DIR / workspace
    else:
        return None


def ensure_overwritable(*paths):
    """
    This is a (nasty) workaround for the permissions issues we hit when
    switching between running the job-runner inside Docker and running it
    natively on Windows. The issue is that the Docker process creates files
    which the Windows native process then doesn't have permission to delete or
    replace. We work around this here by using Docker to delete the offending
    files for us.

    Note for the potentially confused: Windows permissions work nothing like
    POSIX permissions. We can create new files in directories created by
    Docker, we just can't modify or delete existing files.
    """
    if not config.ENABLE_PERMISSIONS_WORKAROUND:
        return
    non_writable = []
    for path in paths:
        path = Path(path)
        if path.exists():
            # It would be nice to have a read-only way of determining if we
            # have write access but I can't seem to find one that works on
            # Windows
            try:
                path.touch()
            except PermissionError:
                non_writable.append(path.resolve())
    if not non_writable:
        return
    root = os.path.commonpath([f.parent for f in non_writable])
    rel_paths = [f.relative_to(root) for f in non_writable]
    rel_posix_paths = [str(f).replace(os.path.sep, "/") for f in rel_paths]
    subprocess_run(
        [
            "docker",
            "run",
            "--rm",
            "--volume",
            f"{root}:/workspace",
            "--workdir",
            "/workspace",
            docker.MANAGEMENT_CONTAINER_IMAGE,
            "rm",
            *rel_posix_paths,
        ],
        check=True,
        capture_output=True,
    )



import argparse
import json


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--config")
    parser.add_argument("output_file")
    args = parser.parse_args()

    config = json.loads(args.config)
    counts = {}
    for input_file in config["files"]:
        with open(input_file) as f:
            counts[input_file] = len(f.readlines())
    with open(args.output_file, "w") as f:
        for name, count in counts.items():
            f.write(f"{name}: {count}\n")



from databuilder.definition import register
from databuilder.query_language import Dataset
from databuilder.tables import patients


year_of_birth = patients.date_of_birth.year
dataset = Dataset()
dataset.set_population(year_of_birth >= 2000)
dataset.year_of_birth = year_of_birth
register(dataset)



import csv
import sys


if __name__ == "__main__":
    sex, input_file, output_file = sys.argv[1:]
    with open(input_file) as f:
        reader = csv.DictReader(f)
        rows = list(reader)
    with open(output_file, "w") as f:
        writer = csv.DictWriter(f, rows[0].keys())
        writer.writeheader()
        for row in rows:
            if row["sex"] == sex:
                writer.writerow(row)



import collections
import csv
import sys


def main(input_file, output_file):
    with open(input_file, newline="") as f_in:
        csv_reader = csv.DictReader(f_in)
        counter = collections.Counter(int(x["year_of_birth"]) for x in csv_reader)

    with open(output_file, "w", newline="") as f_out:
        csv_writer = csv.writer(f_out)
        csv_writer.writerow(["year", "count"])
        csv_writer.writerows(x for x in sorted(counter.items(), key=lambda x: x[0]))


if __name__ == "__main__":
    main(*sys.argv[1:])



"""
Utility functions for interacting with git
"""
import logging
import os
import subprocess
import time
from pathlib import Path, PurePath
from urllib.parse import urlparse, urlunparse

from jobrunner import config
from jobrunner.lib.string_utils import project_name_from_url
from jobrunner.lib.subprocess_utils import subprocess_run


log = logging.getLogger(__name__)


# See `commit_already_fetched`
SENTINEL_TAG_PREFIX = "fetched/"

# Prevent git from ever prompting for credentials. Hat tip:
# https://serverfault.com/a/1054253
NEVER_PROMPT_FOR_AUTH_ENV = dict(
    os.environ,
    GIT_TERMINAL_PROMPT="0",
    GIT_ASKPASS="echo",
    GCM_INTERACTIVE="never",
)


class GitError(Exception):
    pass


class GitFileNotFoundError(GitError):
    pass


class GitRepoNotReachableError(GitError):
    pass


class GitUnknownRefError(GitError):
    pass


def read_file_from_repo(repo_url, commit_sha, path):
    """
    Return the contents of the file at `path` in `repo_url` as of `commit_sha`
    """
    repo_dir = get_local_repo_dir(repo_url)
    ensure_commit_fetched(repo_dir, repo_url, commit_sha)
    try:
        response = subprocess_run(
            ["git", "show", f"{commit_sha}:{path}"],
            capture_output=True,
            check=True,
            cwd=repo_dir,
        )
    except subprocess.SubprocessError as e:
        if e.stderr.startswith(b"fatal: path ") and b"does not exist" in e.stderr:
            raise GitFileNotFoundError(f"File '{path}' not found in repository")
        else:
            log.exception(f"Error reading from {repo_url} @ {commit_sha}")
            raise GitError(f"Error reading from {repo_url} @ {commit_sha}")
    # Note the response here is bytes not text as git doesn't know what
    # encoding the file is supposed to have
    return response.stdout


def checkout_commit(repo_url, commit_sha, target_dir):
    """
    Checkout the contents of `repo_url` as of `commit_sha` into `target_dir`
    """
    repo_dir = get_local_repo_dir(repo_url)
    ensure_commit_fetched(repo_dir, repo_url, commit_sha)
    os.makedirs(target_dir, exist_ok=True)
    subprocess_run(
        [
            "git",
            f"--work-tree={target_dir}",
            "checkout",
            "--quiet",
            "--force",
            commit_sha,
        ],
        check=True,
        # Set GIT_DIR rather than changing working directory so that
        # `target_dir` gets correctly resolved
        env=dict(os.environ, GIT_DIR=repo_dir),
    )


def commit_reachable_from_ref(repo_url, commit_sha, ref):
    """
    Given a `ref` (branch name, tag, etc) on a remote repo, check whether the
    supplied commit is reachable from that ref.
    """
    ref_sha = get_sha_from_remote_ref(repo_url, ref)
    # The easy case and the case I expect to be hit almost every time as the UI
    # currently only supports running against the branch head
    if commit_sha == ref_sha:
        return True
    # However a well (or badly) timed push could cause the target sha and the
    # branch head to diverge, so we need to handle that. In order to do so we
    # need to fetch the history of the branch. We first fetch just the last 10
    # commits on the assumption that it's probably one of those. If that fails
    # we fetch the entire branch history.
    repo_dir = get_local_repo_dir(repo_url)
    ensure_git_init(repo_dir)
    fetch_commit(repo_dir, repo_url, ref_sha, depth=10)
    if commit_is_ancestor(repo_dir, commit_sha, ref_sha):
        return True
    # The below is a git magic number meaning "infinite depth". See:
    # https://git-scm.com/docs/shallow
    fetch_commit(repo_dir, repo_url, ref_sha, depth=2147483647)
    return commit_is_ancestor(repo_dir, commit_sha, ref_sha)


def get_sha_from_remote_ref(repo_url, ref):
    """Gets the SHA of the commit associated with the ref at the repo URL.

    Args:
        repo_url: A repo URL.
        ref: A ref, such as a branch name, tag name, etc.

    Returns:
        The SHA of the commit. For example, if the ref is an annotated tag, then the SHA
        will be that of the associated commit, rather than that of the annotated tag.

    Raises:
        GitRepoNotReachableError: We couldn't read from the remote repo
        GitUnknownRefError: We couldn't find the specified ref in the remote repo
    """
    # If `ref` matches an annotated tag, then `deref_ref` will match the associated
    # commit.
    deref_ref = f"{ref}^{{}}"
    try:
        response = subprocess_run(
            [
                "git",
                "ls-remote",
                "--quiet",
                add_access_token_and_proxy(repo_url),
                ref,
                deref_ref,
            ],
            check=True,
            capture_output=True,
            text=True,
            encoding="utf-8",
            env=NEVER_PROMPT_FOR_AUTH_ENV,
        )
        output = response.stdout
    except subprocess.SubprocessError as exc:
        redact_token_from_exception(exc)
        log.exception("Error reading from remote repository")
        raise GitRepoNotReachableError(f"Could not read from {repo_url}")
    results = _parse_ls_remote_output(output)
    for target_ref in [
        ref,  # e.g. HEAD
        f"refs/heads/{ref}",  # Branch
        f"refs/tags/{deref_ref}",  # Annotated tag
        f"refs/tags/{ref}",  # Lightweight tag
    ]:
        if target_ref in results:
            return results[target_ref]
    raise GitUnknownRefError(f"Could not find ref '{ref}' in {repo_url}")


def _parse_ls_remote_output(output):
    lines = [line.split() for line in output.splitlines()]
    return {line[1]: line[0] for line in lines}


def get_local_repo_dir(repo_url):
    # We don't need to worry that repo_name may not be unique here (e.g. if we
    # end up using repos from different organisations): we're just treating
    # these directories as big buckets of commits, so we could in principle use
    # the same local git directory for all repositories and it would work fine.
    # But it's probably more operationally convenient to split them up like
    # this.
    repo_name = project_name_from_url(repo_url)
    return config.GIT_REPO_DIR / Path(repo_name).with_suffix(".git")


def ensure_commit_fetched(repo_dir, repo_url, commit_sha):
    ensure_git_init(repo_dir)
    # It's safe to keep re-fetching the same commit, but it requires
    # talking to the remote repo every time so it's better to avoid it if
    # we can
    if not commit_already_fetched(repo_dir, commit_sha):
        fetch_commit(repo_dir, repo_url, commit_sha)


def ensure_git_init(repo_dir):
    if not os.path.exists(repo_dir / "config"):
        subprocess_run(["git", "init", "--bare", "--quiet", repo_dir], check=True)


def commit_already_fetched(repo_dir, commit_sha):
    """
    Return whether a given commit exists in a repo directory

    We used to do this with:

        git cat-file -e 'COMMIT_SHA^{commit}'

    However it's possible that an interrupted fetch leaves the commit object in
    place without all of its associated blobs, meaning that the above check
    passes but attempting to check out the commit will fail. To work around
    this we create a special "sentinel" tag for each commit to indicate that
    the entire fetch process has completed successfully.
    """
    response = subprocess_run(
        [
            "git",
            "tag",
            "--list",
            SENTINEL_TAG_PREFIX + commit_sha,
            "--points-at",
            commit_sha,
            "--format",
            "exists",
        ],
        check=True,
        capture_output=True,
        cwd=repo_dir,
    )
    return response.stdout.strip() == b"exists"


def mark_commmit_as_fetched(repo_dir, commit_sha):
    """
    Create a special "sentinel" tag to indicate that the supplied commit has
    been fully fetched (see `commit_already_fetched` above)
    """
    subprocess_run(
        [
            "git",
            "tag",
            "--force",
            SENTINEL_TAG_PREFIX + commit_sha,
            commit_sha,
        ],
        check=True,
        capture_output=True,
        cwd=repo_dir,
    )


def fetch_commit(repo_dir, repo_url, commit_sha, depth=1):
    # The unfortunate retry complexity here is due to mysterious errors we
    # sometimes get when fetching commits in the live environment:
    #
    #   error: RPC failed; curl 56 GnuTLS recv error (-9): A TLS packed with unexpected length was received
    #
    # They are mostly transient (hence the retries) but certain commits seem to
    # trigger them more often than others so presumably it's something to do
    # with the precise sequence of packets that get sent. More details here:
    # https://github.com/opensafely/job-runner/issues/5
    max_retries = 5
    sleep = 4
    attempt = 1
    authenticated_url = add_access_token_and_proxy(repo_url)
    while True:
        try:
            subprocess_run(
                [
                    "git",
                    "fetch",
                    "--force",
                    "--depth",
                    str(depth),
                    authenticated_url,
                    commit_sha,
                ],
                check=True,
                capture_output=True,
                cwd=repo_dir,
                env=NEVER_PROMPT_FOR_AUTH_ENV,
            )
            mark_commmit_as_fetched(repo_dir, commit_sha)
            break
        except subprocess.SubprocessError as e:
            redact_token_from_exception(e)
            log.exception(f"Error fetching commit (attempt {attempt}/{max_retries})")
            if (
                b"GnuTLS recv error" in e.stderr
                or b"SSL_read: Connection was reset" in e.stderr
            ):
                attempt += 1
                if attempt > max_retries:
                    raise GitError(
                        f"Network error when fetching commit {commit_sha} from"
                        f" {repo_url}\n"
                        "(This may work if you try again later)"
                    )
                else:
                    time.sleep(sleep)
                    sleep *= 2
            else:
                raise GitError(f"Error fetching commit {commit_sha} from {repo_url}")


def commit_is_ancestor(repo_dir, ancestor_sha, descendant_sha):
    response = subprocess_run(
        ["git", "merge-base", "--is-ancestor", ancestor_sha, descendant_sha],
        cwd=repo_dir,
        capture_output=True,
    )
    return response.returncode == 0


def add_access_token_and_proxy(repo_url):
    # We've already validated that the repo url starts with https://github.com
    repo_url = repo_url.replace("github.com", config.GIT_PROXY_DOMAIN)
    # We previously did a complicated thing involving the GIT_ASKPASS
    # executable which worked OK on Linux but not on Windows or macOS, so we're
    # doing the more reliable thing of just sticking the token in the URL
    token = config.PRIVATE_REPO_ACCESS_TOKEN
    if not token:
        return repo_url
    # Ensure we only ever send our token to github.com over https
    parsed = urlparse(repo_url)
    if parsed.hostname != config.GIT_PROXY_DOMAIN or parsed.scheme != "https":
        return repo_url
    # Don't overwrite existing auth details (not sure why they'd be there but
    # seems polite)
    if parsed.username or parsed.password:
        return repo_url
    # Github accepts arbitrary usernames when using a PAT so this is just for
    # easy identification in the proxy logs
    username = f"jobrunner-{config.BACKEND}"
    # Add the token to the URL
    return urlunparse(parsed._replace(netloc=f"{username}:{token}@{parsed.netloc}"))


def redact_token_from_exception(exception):
    # The disadvantage of the above approach is it we have to do some work to
    # avoid leaking the token into the logs. However, even if it does leak it's
    # not the end of the world: the developers who have access to the logs have
    # access to the token in any case; and all it provides is read-only access
    # to some private git repos which will eventually become public in any
    # case.
    token = config.PRIVATE_REPO_ACCESS_TOKEN
    if not token:
        return
    if isinstance(exception.cmd, list):
        exception.cmd = [redact(arg, token) for arg in exception.cmd]
    else:
        exception.cmd = redact(exception.cmd, token)
    if exception.output is not None:
        exception.output = redact(exception.output, token)
    if exception.stderr is not None:
        exception.stderr = redact(exception.stderr, token)


def redact(value, secret):
    mask = "********"
    if isinstance(value, str):
        return value.replace(secret, mask)
    elif isinstance(value, bytes):
        return value.replace(secret.encode("ascii"), mask.encode("ascii"))
    # subprocess arguments can also be pathlib.Path instances (PurePath is the
    # base class for all platform-specific Path classes). We never put the
    # token in a path so there's nothing to do here
    elif isinstance(value, PurePath):
        return value
    else:
        raise ValueError(f"Got {type(value)} expected str, bytes or Path")



"""
Super-crude ORM layer than works with dataclasses and implements just the bare
minimum of database functions we need. There was some discussion earlier about
avoiding heavyweight external dependencies like SQLAlchemy hence this little
piece of NIH-ism. However, given that we're going to be relying on external
dependencies for YAML parsing it might make sense to replace this with
something like SQLAlchemy, pinned to a known compromise-free version. The API
surface area of this module is sufficiently small that swapping it out
shouldn't be too large a job.
"""
import dataclasses
import json
import sqlite3
import threading
from enum import Enum

from jobrunner import config


CONNECTION_CACHE = threading.local()


def insert(item):
    table = item.__tablename__
    fields = dataclasses.fields(item)
    columns = ", ".join(escape(field.name) for field in fields)
    placeholders = ", ".join(["?"] * len(fields))
    sql = f"INSERT INTO {escape(table)} ({columns}) VALUES({placeholders})"
    get_connection().execute(sql, encode_field_values(fields, item))


def update(item, exclude_fields=None):
    assert item.id
    exclude_fields = exclude_fields or []
    update_fields = [
        f.name for f in dataclasses.fields(item) if f.name not in exclude_fields
    ]
    update_dict = {f: getattr(item, f) for f in update_fields}
    update_where(item.__class__, update_dict, id=item.id)


def update_where(itemclass, update_dict, **query_params):
    table = itemclass.__tablename__
    fields = [f for f in dataclasses.fields(itemclass) if f.name in update_dict]
    assert len(fields) == len(update_dict)
    updates = ", ".join(f"{escape(field.name)} = ?" for field in fields)
    update_params = encode_field_values(fields, update_dict)
    where, where_params = query_params_to_sql(query_params)
    get_connection().execute(
        f"UPDATE {escape(table)} SET {updates} WHERE {where}",
        update_params + where_params,
    )


def find_where(itemclass, **query_params):
    table = itemclass.__tablename__
    fields = dataclasses.fields(itemclass)
    where, params = query_params_to_sql(query_params)
    sql = f"SELECT * FROM {escape(table)} WHERE {where}"
    cursor = get_connection().execute(sql, params)
    return [itemclass(*decode_field_values(fields, row)) for row in cursor]


def find_all(itemclass):
    return find_where(itemclass)


def find_one(itemclass, **query_params):
    results = find_where(itemclass, **query_params)
    if len(results) == 0:
        raise ValueError(
            f"Found no {itemclass.__name__}s matching {query_params}, expecting one"
        )
    if len(results) > 1:
        raise ValueError(
            f"Found {len(results)} {itemclass.__name__}s matching {query_params}, expecting only one"
        )
    return results[0]


def exists_where(itemclass, **query_params):
    table = itemclass.__tablename__
    where, params = query_params_to_sql(query_params)
    sql = f"SELECT EXISTS (SELECT 1 FROM {escape(table)} WHERE {where})"
    cursor = get_connection().execute(sql, params)
    return bool(cursor.fetchone()[0])


def count_where(itemclass, **query_params):
    table = itemclass.__tablename__
    where, params = query_params_to_sql(query_params)
    sql = f"SELECT COUNT(*) FROM {escape(table)} WHERE {where}"
    cursor = get_connection().execute(sql, params)
    return cursor.fetchone()[0]


def select_values(itemclass, column, **query_params):
    table = itemclass.__tablename__
    fields = [f for f in dataclasses.fields(itemclass) if f.name == column]
    assert fields
    where, params = query_params_to_sql(query_params)
    sql = f"SELECT {escape(column)} FROM {escape(table)} WHERE {where}"
    cursor = get_connection().execute(sql, params)
    return [decode_field_values(fields, row)[0] for row in cursor]


def transaction():
    # Connections function as context managers which create transactions.
    # See: https://docs.python.org/3/library/sqlite3.html#using-the-connection-as-a-context-manager
    # We're relying here on the fact that because of the lru_cache,
    # `get_connection` actually returns the same connection instance every time
    conn = get_connection()
    conn.execute("BEGIN")
    return conn


def get_connection():
    # The caching below means we get the same connection to the database every
    # time which is done not so much for efficiency as so that we can easily
    # implement transaction support without having to explicitly pass round a
    # connection object. This is done on a per-thread basis to avoid potential
    # threading issues.
    filename = config.DATABASE_FILE
    # Looks icky but is documented `threading.local` usage
    cache = CONNECTION_CACHE.__dict__
    if filename in cache:
        return cache[filename]
    else:
        connection = get_connection_from_file(filename)
        cache[filename] = connection
        return connection


def get_connection_from_file(filename):
    if str(filename).startswith(":memory:"):
        filename = ":memory:"
    else:
        filename.parent.mkdir(parents=True, exist_ok=True)
    conn = sqlite3.connect(filename)
    # Enable autocommit so changes made outside of a transaction still get
    # persisted to disk. We can use explicit transactions when we need
    # atomicity.
    conn.isolation_level = None
    # Support dict-like access to rows
    conn.row_factory = sqlite3.Row
    schema_count = list(conn.execute("SELECT COUNT(*) FROM sqlite_master"))[0][0]
    if schema_count == 0:
        with open(config.DATABASE_SCHEMA_FILE) as f:
            schema_sql = f.read()
        conn.executescript(schema_sql)
    return conn


def query_params_to_sql(params):
    """
    Turn a dict of query parameters into a pair of (SQL string, SQL values).
    All parameters are implicitly ANDed together, and there's a bit of magic to
    handle `field__in=list_of_values` queries, LIKE queries and Enum classes.
    """
    parts = []
    values = []
    for key, value in params.items():
        if key.endswith("__in"):
            field = key[:-4]
            placeholders = ", ".join(["?"] * len(value))
            parts.append(f"{escape(field)} IN ({placeholders})")
            values.extend(value)
        elif key.endswith("__like"):
            field = key[:-6]
            parts.append(f"{escape(field)} LIKE ?")
            values.append(value)
        else:
            parts.append(f"{escape(key)} = ?")
            values.append(value)
    # Bit of a hack: convert any Enum instances to their values so we can use
    # them in querying
    values = [v.value if isinstance(v, Enum) else v for v in values]
    if not parts:
        parts = ["1 = 1"]
    return " AND ".join(parts), values


def escape(s):
    """
    Escape SQLite identifier (as opposed to string literal)
    See https://www.sqlite.org/lang_keywords.html
    """
    return '"{}"'.format(s.replace('"', '""'))


def encode_field_values(fields, item):
    """
    Takes a list of dataclass fields and a dataclass instance or dict and
    returns the field values as a list with the appropriate conversions applied
    """
    values = []
    get_value = getattr if not isinstance(item, dict) else dict.__getitem__
    for field in fields:
        value = get_value(item, field.name)
        # Dicts and lists get encoded as JSON
        if field.type in (list, dict) and value is not None:
            value = json.dumps(value)
        # Enums get encoded as their string/int values
        elif issubclass(field.type, Enum) and value is not None:
            value = value.value
        values.append(value)
    return values


def decode_field_values(fields, row):
    """
    Takes a list of dataclass fields and a SQLite row (or any dict-like) and
    returns field values as a list with the appropriate conversions applied
    """
    values = []
    for field in fields:
        value = row[field.name]
        # Dicts and lists get decoded from JSON
        if field.type in (list, dict) and value is not None:
            value = json.loads(value)
        # Enums get transformed back from their string/int values
        elif issubclass(field.type, Enum) and value is not None:
            value = field.type(value)
        values.append(value)
    return values



from cohortextractor import StudyDefinition, patients


study = StudyDefinition(
    default_expectations={
        "date": {"earliest": "1970-01-01", "latest": "today"},
        "rate": "uniform",
        "incidence": 0.2,
    },
    population=patients.registered_with_one_practice_between(
        "2019-02-01", "2020-02-01"
    ),
    dereg_date=patients.date_deregistered_from_all_supported_practices(
        on_or_before="2020-12-01",
        date_format="YYYY-MM",
        return_expectations={"date": {"earliest": "2020-02-01"}},
    ),
    a_e_consult_date=patients.attended_emergency_care(
        on_or_after="2020-02-01",
        returning="date_arrived",
        date_format="YYYY-MM-DD",
        find_first_match_in_period=True,
        return_expectations={
            "date": {"earliest": "2020-02-01"},
            "rate": "exponential_increase",
        },
    ),
    icu_date_admitted=patients.admitted_to_icu(
        on_or_after="2020-02-01",
        include_day=True,
        returning="date_admitted",
        find_first_match_in_period=True,
        return_expectations={
            "date": {"earliest": "2020-02-01"},
            "rate": "exponential_increase",
        },
    ),
    died_date_cpns=patients.with_death_recorded_in_cpns(
        on_or_after="2020-02-01",
        returning="date_of_death",
        include_month=True,
        include_day=True,
        return_expectations={
            "date": {"earliest": "2020-02-01"},
            "rate": "exponential_increase",
        },
    ),
    age=patients.age_as_of(
        "2020-02-01",
        return_expectations={
            "rate": "universal",
            "int": {"distribution": "population_ages"},
        },
    ),
    sex=patients.sex(
        return_expectations={
            "rate": "universal",
            "category": {"ratios": {"M": 0.49, "F": 0.51}},
        }
    ),
    imd=patients.address_as_of(
        "2020-02-01",
        returning="index_of_multiple_deprivation",
        round_to_nearest=100,
        return_expectations={
            "rate": "universal",
            "category": {"ratios": {"100": 0.1, "200": 0.2, "300": 0.7}},
        },
    ),
)



import secrets
from contextlib import contextmanager


@contextmanager
def atomic_writer(dest):
    """Return a safe temp file on the same filesystem to write to

    On success, the tmp file is renamed to the original target atomically.
    If the write fails, ensure the tmp file is deleted
    """
    dest.parent.mkdir(parents=True, exist_ok=True)
    tmp = dest.with_suffix(dest.suffix + f".{secrets.token_hex(8)}.tmp")
    try:
        yield tmp
    except Exception:
        tmp.unlink(missing_ok=True)
        raise
    else:
        tmp.replace(dest)



import dataclasses
import shlex
import textwrap

from jobrunner import config
from jobrunner.lib import git
from jobrunner.lib.github_validators import (
    GithubValidationError,
    validate_branch_and_commit,
    validate_repo_url,
)
from jobrunner.lib.yaml_utils import YAMLError, parse_yaml
from jobrunner.project import is_generate_cohort_command


class ReusableActionError(Exception):
    """Represents a study developer-friendly reusable action error.

    We raise this in preference to other, lower-level, errors because there's only so
    much a study developer can do when there's an error with a reusable action.
    """


@dataclasses.dataclass
class ReusableAction:
    repo_url: str
    commit: str
    action_file: bytes


def resolve_reusable_action_references(jobs):
    """
    Accepts a list of Job instances, identifies any which invoke reusable
    actions and modifies them appropriately which means:
        * rewriting their `run` command to use the entrypoint defined by the
          reusable action
        * adding a reference to the reusable action's repo and commit

    Args:
        jobs: list of Job instances

    Returns:
        None - it modifies its arguments in place

    Raises:
        ReusableActionError
    """
    for job in jobs:
        try:
            run_command, repo_url, commit = handle_reusable_action(job.run_command)
        except ReusableActionError as e:
            # Annotate the exception with the context of the action in which it
            # occured
            context = f"{job.action}: {job.run_command.split()[0]}"
            raise ReusableActionError(f"in '{context}' {e}") from e
        job.run_command = run_command
        job.action_repo_url = repo_url
        job.action_commit = commit


def handle_reusable_action(run_command):
    """
    If `run_command` refers to a reusable action then rewrite it appropriately
    and return it along with the repo_url and commit of the reusable action.
    Otherwise return it unchanged with null values for repo_url and commit.

    Args:
        run_command: Action's run command as a string

    Returns: tuple consisting of
        - rewritten_run_command: string
        - resuable_action_repo_url: string or None if not a reusable action
        - reusable_action_commit: string or None if not a reusable action

    Raises:
        ReusableActionError: Something was wrong with the reusable action
    """
    run_args = shlex.split(run_command)
    image, tag = run_args[0].split(":")

    if image in config.ALLOWED_IMAGES:
        # This isn't a reusable action, nothing to do
        return run_command, None, None

    reusable_action = fetch_reusable_action(image, tag)
    new_run_args = apply_reusable_action(run_args, reusable_action)
    new_run_command = shlex.join(new_run_args)
    return new_run_command, reusable_action.repo_url, reusable_action.commit


def fetch_reusable_action(image, tag):
    """
    Fetch all metadata from git needed to apply a reusable action

    Args:
        image: The name of the reusable action
        tag: The specified version of the reusable action

    Returns:
        ReusableAction object, wrapping the repo_url, commit and the contents
        of the `action.yaml` file

    Raises:
        ReusableActionError: An error occurred when accessing the reusable action.
    """
    repo_url = f"{config.ACTIONS_GITHUB_ORG_URL}/{image}"
    try:
        validate_repo_url(repo_url, [config.ACTIONS_GITHUB_ORG])
    except GithubValidationError:
        raise ReusableActionError(f"'{image}' contains invalid characters")

    try:
        # If there's a problem, then it relates to the repository. Maybe the study
        # developer made an error; maybe the reusable action developer made an error.
        commit = git.get_sha_from_remote_ref(repo_url, tag)
    except git.GitRepoNotReachableError:
        raise ReusableActionError(
            f"could not find a repo at {repo_url}\n"
            f"Check that '{image}' is in the list of available actions at "
            f"https://actions.opensafely.org"
        )
    except git.GitUnknownRefError:
        raise ReusableActionError(f"'{tag}' is not a tag listed in {repo_url}/tags")

    # We're planning to give write access to specific external collaborators on
    # specific action repos, but we want to retain final control over what gets
    # deployed. Github's permissions model doesn't (yet) allow us to restrict
    # what tags they can push, but we can restrict access to certain branches.
    # So here we check that the tag refers to a commit which has been merged to
    # main and refuse to run if not. If Github ever supports restricting tag
    # creation then we can do away with this check.
    try:
        validate_branch_and_commit(repo_url, commit, "main")
    except GithubValidationError:
        raise ReusableActionError(
            f"tag '{tag}' has not yet been approved for use (not merged into main branch)"
        )
    except git.GitError:
        # Our git library already logs the relevant details here so throwing
        # away the original exception is fine
        raise ReusableActionError(f"error validating '{commit}' in {repo_url}")

    try:
        # If there's a problem, then it relates to the reusable action. The study
        # developer didn't make an error; the reusable action developer did.
        action_file = git.read_file_from_repo(repo_url, commit, "action.yaml")
    except git.GitFileNotFoundError:
        raise ReusableActionError(
            f"{repo_url}/tree/{tag} doesn't look like a valid action "
            f"(no 'action.yaml' file present)"
        )
    except git.GitError:
        # Our git library already logs the relevant details here so throwing
        # away the original exception is fine
        raise ReusableActionError(f"error reading '{commit}' from {repo_url}")

    return ReusableAction(repo_url=repo_url, commit=commit, action_file=action_file)


def apply_reusable_action(run_args, reusable_action):
    """
    Rewrite a list of "run" arguments to run the code specifed by the supplied
    `ReusableAction` instance.

    Args:
        run_args: Action's run command as a list of string arguments
        reusable_action: A ReusableAction instance

    Returns:
        The modified run arguments as a list

    Raises:
        ReusableActionError: An error occurred when accessing the reusable action.
    """
    try:
        # If there's a problem, then it relates to the reusable action. The study
        # developer didn't make an error; the reusable action developer did.
        action_config = parse_yaml(reusable_action.action_file, name="action.yaml")
        if "run" not in action_config:
            raise ReusableActionError("Missing `run` key in 'action.yaml'")
        action_run_args = shlex.split(action_config["run"])
        action_image, action_tag = action_run_args[0].split(":")
        if action_image not in config.ALLOWED_IMAGES:
            raise ReusableActionError(f"Unrecognised runtime: {action_image}")
        if is_generate_cohort_command(action_run_args):
            raise ReusableActionError(
                "Re-usable actions cannot invoke cohortextractor/databuilder"
            )
    except (YAMLError, ReusableActionError) as e:
        formatted_error = textwrap.indent(f"{type(e).__name__}: {e}", "  ")
        raise ReusableActionError(
            f"invalid action, please open an issue on "
            f"{reusable_action.repo_url}/issues\n\n"
            f"{formatted_error}"
        )

    # ["action:tag", "arg", ...] -> ["runtime:tag binary entrypoint", "arg", ...]
    return action_run_args + run_args[1:]



"""
Tool for migrating data from manifests to the database.

Usage:

    python -m jobrunner.cli.manifest_migration --batch-size 50

Reads job records from manifest files in all workspaces and creates records in the database for any that are missing.
For jobs that already exist in the database, it doesn't make any attempt to check that they are consistent with the
record in the manifest.

The number of jobs that are created list limited by the --batch-size argument. If there are more jobs that could have
been migrated then a message is displayed when the tool exits. The operator is expected to run the tool once with a
small batch size to check that it's working correctly and then use a large batch size to complete the operation. The
tool will report when there are no further jobs to be migrated.
"""
import argparse
import sys

from jobrunner import manifest_to_database_migration


def main(args=None):
    args = args or []
    parser = argparse.ArgumentParser(description=__doc__.partition("\n\n")[0])
    parser.add_argument(
        "--batch-size", type=int, default=1, help="maximum number of jobs to create"
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="do a dry run migration without modifying data and carrying on if there is an error with a single job or workspace",
    )
    parsed = parser.parse_args(args)

    manifest_to_database_migration.migrate_all(
        batch_size=parsed.batch_size,
        dry_run=parsed.dry_run,
        ignore_errors=parsed.dry_run,
    )


if __name__ == "__main__":
    main(sys.argv[1:])



import time

import pytest

from jobrunner.lib import docker
from jobrunner.lib.docker_stats import (
    get_container_stats,
    get_volume_and_container_sizes,
)


@pytest.mark.needs_docker
# This runs fine locally but fails in CI, despite the retry logic added below.
# Don't have time to diagnose this properly and this isn't core functionality
# in any case
@pytest.mark.xfail
def test_get_container_stats(docker_cleanup):
    docker.run("test_container1", [docker.MANAGEMENT_CONTAINER_IMAGE, "sh"])
    # It can sometimes take a while before the container actually appears :(
    for _ in range(10):
        containers = get_container_stats()
        if "test_container1" not in containers:
            time.sleep(1)
        else:
            break
    assert containers["test_container1"] == {"cpu_percentage": 0, "memory_used": 0}


@pytest.mark.needs_docker
@pytest.mark.slow_test
def test_get_volume_and_container_sizes(tmp_path, docker_cleanup):
    half_meg_file = tmp_path / "halfmeg"
    half_meg_file.write_bytes(b"0" * 500000)
    docker.create_volume("test_volume1")
    docker.copy_to_volume("test_volume1", half_meg_file, "halfmeg")
    docker.run(
        "test_container2",
        [docker.MANAGEMENT_CONTAINER_IMAGE, "cp", "/workspace/halfmeg", "/halfmeg"],
        volume=("test_volume1", "/workspace"),
    )
    volumes, containers = get_volume_and_container_sizes()
    assert volumes["test_volume1"] == 500000
    assert containers["test_container2"] == 500000



import collections


class LRUDict(collections.OrderedDict):
    def __init__(self, capacity):
        super().__init__()
        self._capacity = capacity

    def __getitem__(self, key):
        value = super().__getitem__(key)
        self.move_to_end(key)
        return value

    def get(self, key, default=None):
        # We have to provide our own implementation to ensure that it delegates to our modified __getitem__().
        if key in self:
            return self.__getitem__(key)
        return None

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.move_to_end(key)
        if len(self) > self._capacity:
            self._remove_lru()

    def _remove_lru(self):
        # We can't use self.popitem() as it calls __getitem__().
        del self[next(iter(self))]



import logging
from typing import Callable, List

from jobrunner.job_executor import (
    ExecutorAPI,
    JobDefinition,
    JobResults,
    JobStatus,
    Privacy,
)
from jobrunner.lib.lru_dict import LRUDict


LOGGER_NAME = "executor"


class LoggingExecutor(ExecutorAPI):
    def __init__(self, wrapped: ExecutorAPI):
        self._logger = logging.getLogger(LOGGER_NAME)
        self._state_cache = LRUDict(100)  # Maps job ids to states
        self._wrapped = wrapped
        self._add_logging(self._wrapped.get_status)
        self._add_logging(self._wrapped.prepare)
        self._add_logging(self._wrapped.execute)
        self._add_logging(self._wrapped.finalize)
        self._add_logging(self._wrapped.terminate)
        self._add_logging(self._wrapped.cleanup)

    def get_results(self, job: JobDefinition) -> JobResults:
        return self._wrapped.get_results(job)

    def delete_files(self, workspace: str, privacy: Privacy, paths: [str]) -> List[str]:
        return self._wrapped.delete_files(workspace, privacy, paths)

    def _add_logging(self, method: Callable[[JobDefinition], JobStatus]):
        def wrapper(job: JobDefinition) -> JobStatus:
            status = method(job)
            if self._is_new_state(job, status.state):
                self._write_log(job, status)
                self._state_cache[job.id] = status.state
            return status

        setattr(self, method.__name__, wrapper)

    def _is_new_state(self, job, state):
        return job.id not in self._state_cache or self._state_cache[job.id] != state

    def _write_log(self, job, status):
        log = f"State change for job {job.id}: {self._state_cache.get(job.id)} -> {status.state}"
        if status.message:
            log += f" ({status.message})"
        self._logger.info(log)



import re
from urllib.parse import urlparse


def slugify(s):
    s = s.encode("ascii", "ignore").decode("ascii").lower()
    # Remove anything that's not alphanumeric, underscore, dash or whitespace
    s = re.sub(r"[^\w\s\-]", "", s)
    # Replace repeated runs of anything non-alphanumeric with a single dash
    return re.sub(r"[_\s\-]+", "-", s).strip("-")


def project_name_from_url(url):
    """
    Return the name of a repository from its URL (there's nothing particularly
    significant about a repository's name but it can make debugging easier to
    include it in various places)
    """
    # "URL" here can in fact be a local path and it may be a Windows path, in
    # which case we need to convert the slash type so that it gets handled
    # correctly
    url = url.replace("\\", "/")
    name = urlparse(url).path.strip("/").split("/")[-1]
    if name.endswith(".git"):
        name = name[:-4]
    return name


def tabulate(rows, separator=" ", indent=0, empty=""):
    """
    Formats two columns of data
    """
    if not rows:
        return f"{' ' * indent}{empty}"
    max_col_0 = max(len(row[0]) for row in rows)
    max_col_1 = max(len(row[1]) for row in rows)
    format_str = f"{' ' * indent}{{0:<{max_col_0}}}{separator}{{1:<{max_col_1}}}"
    return "\n".join(format_str.format(*row) for row in rows)



from dataclasses import dataclass
from enum import Enum
from typing import List, Mapping, Optional


class Privacy(Enum):
    HIGH = "high"
    MEDIUM = "medium"


@dataclass
class Study:
    git_repo_url: str
    commit: str


@dataclass
class JobDefinition:
    id: str  # a unique identifier for the job  # noqa: A003
    study: Study  # the study defining the action for this job
    workspace: str  # the workspace to run the job in
    action: str  # the name of the action that the job is running
    image: str  # the Docker image to run
    args: List[str]  # the arguments to pass to the Docker container
    env: Mapping[str, str]  # the environment variables to set for the Docker container
    inputs: List[str]  # the files that the job requires
    output_spec: Mapping[
        str, str
    ]  # the files that the job should produce (globs mapped to privacy levels)
    allow_database_access: bool  # whether this job should have access to the database


class ExecutorState(Enum):
    PREPARING = "preparing"
    PREPARED = "prepared"
    EXECUTING = "executing"
    EXECUTED = "executed"
    FINALIZING = "finalizing"
    FINALIZED = "finalized"
    UNKNOWN = "unknown"
    ERROR = "error"


@dataclass
class JobStatus:
    state: ExecutorState
    message: Optional[str] = None


@dataclass
class JobResults:
    outputs: Mapping[str, str]  # mapping of outputs to privacy levels
    unmatched_patterns: List[str]  # list of patterns that matched no outputs
    exit_code: int
    image_id: str
    message: str = None


class ExecutorAPI:
    """
    API for managing job execution.

    This API is called by the job-runner to manage each job it is tracking. It models the running of a job as a state
    machine, and the methods below are the transitions between states.

    They should return either:
     - the next state if successful
     - the ERROR state with an appropriate message if something has gone wrong.
     - the current state if is different from the expected state, with a message
     - the initial state to indicate back pressure and to retry later.

    Given the long running nature of jobs, it is an asynchronous API, and calls should not block for a more than a few
    seconds.

    All the state transition methods (prepare(), execute(), finalize(), terminate(), cleanup()) must be idempotent. If
    the relevant task they are responsible for is already running for that job, they must not start a new task, and
    instead return successfully the current state. This is best acheived by a guard to check the job's state, and
    return its current state if in the state you would expect. It's the responsibility of the job-runner scheduler to
    figure out what to do in this case.

    """

    def prepare(self, job: JobDefinition) -> JobStatus:
        """
        Launch a prepare task for a job, transitioning to the initial PREPARING state.

        1. Validate the JobDefinition. If there are errors, return an ERROR state with message.

        2. Check the job is currently in UNKNOWN state. If not return its current state, and the job-runner will handle
           the unexpected transition.

        3. Check the resources are available to prepare the job. If not, return the UNKNOWN state with an appropriate
           message.

        4. Create an ephemeral workspace to use for executing this job. This is expected to be a volume mounted into the
           container, but other implementations are allowed.

        5. Launch a prepare task asynchronously. If launched successfully, return the PREPARING state. If not, return an
           ERROR state with message.

        The prepare task must do the following:

          - check out the supplied study repo via the OpenSAFELY github proxy into the ephemeral workspace, erroring if
            there are any failures.
          - copying the supplied file inputs from the long-term workspace storage into the ephemeral workspace, erroring
            if there are any missing.

        When the prepare task finishes, the get_status() call should now return PREPARED for this job.

        This method must be idempotent. If called with a job that is already running a prepare task, it must not
        launch a new task, and simply return successfully with PREPARING.

        """

    def execute(self, job: JobDefinition) -> JobStatus:
        """
        Launch the execution of a job that has been prepared, transitioning from PREPARED to EXECUTING.

        1. Check the job is in the PREPARED state. If not, return its current state, and the job-runner will handle the
           unexpected transition.

        2. Validate that the ephemeral workspace created by prepare for this job exists.  If not, return an ERROR
           state with message.

        3. Check there are resources available to execute the job. If not, return PREPARED status with an appropriate
           message.

        4. Launch the job execution task asynchronously. If launched successfully, return the EXECUTING state. If not,
           return an ERROR state with message.

        The execution task must do the following:

        The specified image must be run, with the provided arguments and environment variables. The implementation
        may add environment variables to those in the job definition as necessary for the backend.

        The job should be run without any network access, unless definition.allow_database_access is set to True,
        in which case it should be run with a network allowing access to the database and any configuration needed to
        contact and authenticate with the database should be provided as environment variables.

        The job must be run with the ephemeral workspace for this job at /workspace in the filesystem.

        When the execute task finishes, the get_status() call must now return EXECUTED for this job.

        This method must be idempotent. If called with a job that is already running an execute task, it must not
        launch a new task, and simply return successfully with EXECUTING.

        """

    def finalize(self, job: JobDefinition) -> JobStatus:
        """
        Launch the finalization of a job, transitioning from EXECUTED to FINALIZING.

        1. Check the job is in the EXECUTED state. If not, return its current state, and the job-runner will handle the
           unexpected transition.

        2. Validate that the job's ephemeral workspace exists. If not, return an ERROR state with message.

        3. Launch the finalize task asynchronously. If launched successfully, return the FINALIZING state. If not,
           return an ERROR state with message.

        The finalize task should do the following:

        Any files that the job produced that match the output spec in the definition must be copied from the ephemeral
        workspace to the workspace long-term storage. Anything written by the container to stdout or stderr must be
        captured and written to a log file, metadata/{action}.log, in the workspace in long-term storage.

        The action log file and any files in the output spec marked as medium privacy must also be made available in the
        medium privacy view of the workspace in long-term storage.

        The action log file and any useful metadata from the job run should also be written to a separate log storage
        area in long-term storage.

        When the finalize task finishes, the get_status() call should now return FINALIZED for this job, and
        get_results() call should return the JobResults for this job.

        This method must be idempotent. If called with a job that is already running an finalize task, it must not
        launch a new task, and simply return successfully with FINALIZING.

        """

    def terminate(self, job: JobDefinition) -> JobStatus:
        """
        Terminate a running job, transitioning to the ERROR state.

        1. If any task for this job is running, terminate it, do not wait for it to complete.

        2. Return ERROR state with a message.

        """

    def cleanup(self, job: JobDefinition) -> JobStatus:
        """
        Clean up any remaining state for a finished job, transitioning to the UNKNOWN state.

        1. Initiate the cleanup, do not wait for it to complete.

        2. Return the UNKNOWN status.

        This method must be idempotent; it will be called at least once for every finished job. The implementation
        may defer resource cleanup to this method if necessary in order to correctly implement idempotency of
        get_status() or get_results(). If the job is unknown, it should still return UNKNOWN successfully.

        This method will not be called for a job that raises an unexpected exception from ExecutorAPI in order
        to facilitate debugging of unexpected failures. It may therefore be necessary for the backend to provide
        out-of-band mechanisms for cleaning up resources associated with such failures.
        """

    def get_status(self, job: JobDefinition) -> JobStatus:
        """
        Return the current status of a job.

        1. Check the job is known to the system. If not, return the UNKNOWN state.

        2. Return the current state of the job from the executors perspective.

        This should return a JobStatus with the appropriate state for the job. It is polled by job-runner to track the
        completion of the various tasks.

        This method must be idempotent; it may be called more than once for a job even after it has finished, so any
        irreversible cleanup which loses information about must be deferred to ExecutorAPI.cleanup() which will only be
        called once the results have been persisted.

        """

    def get_results(self, job: JobDefinition) -> JobResults:
        """
        Return the finalized results for a job.

        The results must include a list of output files that the job produced which matched its output spec. It
        should also include a list of files that it produced but which did not match the output spec, to aid in
        debugging during study development.

        This method must be idempotent; it may be called more than once for a job even after it has finished, so any
        irreversible cleanup which loses information about must be deferred to ExecutorAPI.cleanup() which will only be
        called once the results have been persisted.
        """

    def delete_files(self, workspace: str, privacy: Privacy, paths: [str]) -> List[str]:
        """
        Delete files from a workspace.

        This method must be idempotent; if any of the files specified doesn't exist then it must ignore them.

        Returns a list of any files that were present but it errored trying to delete them.
        """
        ...


class NullExecutorAPI(ExecutorAPI):
    """Null implementation of ExecutorAPI."""

    def prepare(self, job):
        raise NotImplementedError

    def execute(self, job):
        raise NotImplementedError

    def finalize(self, job):
        raise NotImplementedError

    def terminate(self, job):
        raise NotImplementedError

    def get_status(self, job):
        raise NotImplementedError

    def get_results(self, job):
        raise NotImplementedError

    def cleanup(self, job):
        raise NotImplementedError

    def delete_files(self, workspace, privacy, paths):
        raise NotImplementedError



from jobrunner.lib.string_utils import project_name_from_url


def test_project_name_from_url():
    assert project_name_from_url("https://github.com/opensafely/test1.git") == "test1"
    assert project_name_from_url("https://github.com/opensafely/test2/") == "test2"
    assert project_name_from_url("/some/local/path/test3/") == "test3"
    assert project_name_from_url("C:\\some\\windows\\path\\test4\\") == "test4"



import pytest

from jobrunner.lib.database import find_one, insert, select_values, update
from jobrunner.models import Job, State


def test_basic_roundtrip(tmp_work_dir):
    job = Job(
        id="foo123",
        job_request_id="bar123",
        state=State.RUNNING,
        output_spec={"hello": [1, 2, 3]},
    )
    insert(job)
    j = find_one(Job, job_request_id__in=["bar123", "baz123"])
    assert job.id == j.id
    assert job.output_spec == j.output_spec


def test_update(tmp_work_dir):
    job = Job(id="foo123", action="foo")
    insert(job)
    job.action = "bar"
    update(job)
    assert find_one(Job, id="foo123").action == "bar"


def test_update_excluding_a_field(tmp_work_dir):
    job = Job(id="foo123", action="foo", commit="commit-of-glory")
    insert(job)
    job.action = "bar"
    job.commit = "commit-of-doom"
    update(job, exclude_fields=["commit"])
    j = find_one(Job, id="foo123")
    assert j.action == "bar"
    assert j.commit == "commit-of-glory"


def test_select_values(tmp_work_dir):
    insert(Job(id="foo123", state=State.PENDING))
    insert(Job(id="foo124", state=State.RUNNING))
    insert(Job(id="foo125", state=State.FAILED))
    values = select_values(Job, "id", state__in=[State.PENDING, State.FAILED])
    assert sorted(values) == ["foo123", "foo125"]
    values = select_values(Job, "state", id="foo124")
    assert values == [State.RUNNING]


def test_find_one_returns_a_single_value(tmp_work_dir):
    insert(Job(id="foo123", workspace="the-workspace"))
    job = find_one(Job, id="foo123")
    assert job.workspace == "the-workspace"


def test_find_one_fails_if_there_are_no_results(tmp_work_dir):
    with pytest.raises(ValueError):
        find_one(Job, id="foo123")


def test_find_one_fails_if_there_is_more_than_one_result(tmp_work_dir):
    insert(Job(id="foo123", workspace="the-workspace"))
    insert(Job(id="foo456", workspace="the-workspace"))
    with pytest.raises(ValueError):
        find_one(Job, workspace="the-workspace")



import os
from pathlib import Path

import pytest

from jobrunner.lib.git import (
    GitRepoNotReachableError,
    GitUnknownRefError,
    checkout_commit,
    commit_already_fetched,
    commit_reachable_from_ref,
    ensure_git_init,
    fetch_commit,
    get_sha_from_remote_ref,
    read_file_from_repo,
)


REPO_FIXTURE = str(Path(__file__).parents[1].resolve() / "fixtures/git-repo")


@pytest.mark.slow_test
def test_read_file_from_repo(tmp_work_dir):
    output = read_file_from_repo(
        "https://github.com/opensafely-core/test-public-repository.git",
        "c1ef0e676ec448b0a49e0073db364f36f6d6d078",
        "README.md",
    )
    assert output == b"# test-public-repository"


@pytest.mark.slow_test
def test_checkout_commit(tmp_work_dir, tmp_path):
    target_dir = tmp_path / "files"
    checkout_commit(
        "https://github.com/opensafely-core/test-public-repository.git",
        "c1ef0e676ec448b0a49e0073db364f36f6d6d078",
        target_dir,
    )
    assert [f.name for f in target_dir.iterdir()] == ["README.md"]


@pytest.mark.slow_test
def test_get_sha_from_remote_ref(tmp_work_dir):
    sha = get_sha_from_remote_ref(
        "https://github.com/opensafely-core/test-public-repository.git",
        "test-tag-dont-delete",
    )
    assert sha == "029a6ff81cb0ab878de24c12bc690969163c5c9e"


@pytest.mark.slow_test
def test_get_sha_from_remote_ref_annotated_tag():
    sha = get_sha_from_remote_ref(
        "https://github.com/opensafely-core/test-public-repository.git",
        "test-annotated-tag-dont-delete",
    )
    assert sha == "3c15ff525001e039d4e27cfc62f652ecad09fde4"


@pytest.mark.slow_test
def test_get_sha_from_remote_ref_missing_ref(tmp_work_dir):
    with pytest.raises(GitUnknownRefError):
        get_sha_from_remote_ref(
            "https://github.com/opensafely-core/test-public-repository.git",
            "no-such-ref",
        )


@pytest.mark.slow_test
def test_get_sha_from_remote_ref_missing_repo(tmp_work_dir):
    with pytest.raises(GitRepoNotReachableError):
        get_sha_from_remote_ref(
            "https://github.com/opensafely-core/no-such-repo.git", "main"
        )


@pytest.mark.slow_test
def test_commit_reachable_from_ref(tmp_work_dir):
    is_reachable_good = commit_reachable_from_ref(
        "https://github.com/opensafely-core/test-public-repository.git",
        "029a6ff81cb0ab878de24c12bc690969163c5c9e",
        "test-branch-dont-delete",
    )
    assert is_reachable_good
    is_reachable_bad = commit_reachable_from_ref(
        "https://github.com/opensafely-core/test-public-repository.git",
        "029a6ff81cb0ab878de24c12bc690969163c5c9e",
        "main",
    )
    assert not is_reachable_bad


# These tests makes request to an actual private GitHub repo and so will only
# work if there's an appropriate access token in the environment


@pytest.mark.skipif(
    not os.environ.get("PRIVATE_REPO_ACCESS_TOKEN"),
    reason="No access token in environment",
)
@pytest.mark.slow_test
def test_read_file_from_private_repo(tmp_work_dir):
    output = read_file_from_repo(
        "https://github.com/opensafely/test-repository.git",
        "d7fe87ab5d6dc97222c4a9dbf7c0fe40fc108c8f",
        "README.md",
    )
    assert output == b"# test-repository\nTesting GH permssions model\n"


@pytest.mark.skipif(
    not os.environ.get("PRIVATE_REPO_ACCESS_TOKEN"),
    reason="No access token in environment",
)
@pytest.mark.slow_test
def test_get_sha_from_remote_ref_private(tmp_work_dir):
    sha = get_sha_from_remote_ref(
        "https://github.com/opensafely/test-repository", "v1.0"
    )
    assert sha == "981ac62ec5620df90556bc18784f06b6e7db7e4d"


# The below tests use a local git repo fixture rather than accessing GitHub
# over HTTPS. This makes them faster, though obviously less complete.


def test_read_file_from_repo_local(tmp_work_dir):
    output = read_file_from_repo(
        REPO_FIXTURE,
        "d1e88b31cbe8f67c58f938adb5ee500d54a69764",
        "project.yaml",
    )
    assert output.startswith(b"version: '1.0'")


def test_checkout_commit_local(tmp_work_dir, tmp_path):
    target_dir = tmp_path / "files"
    checkout_commit(
        REPO_FIXTURE,
        "d1e88b31cbe8f67c58f938adb5ee500d54a69764",
        target_dir,
    )
    assert [f.name for f in target_dir.iterdir()] == ["project.yaml"]


def test_get_sha_from_remote_ref_local(tmp_work_dir):
    sha = get_sha_from_remote_ref(REPO_FIXTURE, "v1")
    assert sha == "d1e88b31cbe8f67c58f938adb5ee500d54a69764"


def test_get_sha_from_remote_ref_local_missing_ref(tmp_work_dir):
    with pytest.raises(GitUnknownRefError):
        get_sha_from_remote_ref(REPO_FIXTURE, "no-such-ref")


def test_get_sha_from_remote_ref_local_missing_repo(tmp_work_dir):
    MISSING_REPO = REPO_FIXTURE + "-no-such-repo"
    with pytest.raises(GitRepoNotReachableError):
        get_sha_from_remote_ref(MISSING_REPO, "v1")


def test_commit_already_fetched(tmp_path):
    commit_sha = "d1e88b31cbe8f67c58f938adb5ee500d54a69764"
    repo_dir = tmp_path / "repo"
    ensure_git_init(repo_dir)
    assert not commit_already_fetched(repo_dir, commit_sha)
    fetch_commit(repo_dir, REPO_FIXTURE, commit_sha)
    assert commit_already_fetched(repo_dir, commit_sha)



import importlib

from jobrunner import config
from jobrunner.executors.logging import LoggingExecutor


def get_executor_api():
    module_name, cls = config.EXECUTOR.split(":", 1)
    module = importlib.import_module(module_name)
    return LoggingExecutor(getattr(module, cls)())



"""
Thin wrapper around `subprocess.run` which ensures that any arguments which are
pathlib instances get coerced to strings, which is necessary for them to work
on Windows (but not POSIX). Most of these issues are fixed in Python 3.8 so
it's possible we can drop this later. (The exception being the `env` argument
which the documentation doesn't mention so we'll have to wait and see.)
"""
import subprocess
from pathlib import PurePath


def subprocess_run(cmd_args, **kwargs):
    assert not kwargs.get("shell"), "Don't use shell as we need to work cross-platform"
    cmd_args = list(map(to_str, cmd_args))
    if "cwd" in kwargs:
        kwargs["cwd"] = to_str(kwargs["cwd"])
    if "env" in kwargs:
        kwargs["env"] = {key: to_str(value) for (key, value) in kwargs["env"].items()}
    return subprocess.run(cmd_args, **kwargs)


def to_str(value):
    # PurePath is the base class for all pathlib classes
    if isinstance(value, PurePath):
        return str(value)
    return value



import pytest

from jobrunner.lib import atomic_writer


def test_atomic_writer_success(tmp_path):
    dst = tmp_path / "dst"
    with atomic_writer(dst) as tmp:
        tmp.write_text("dst")

    assert dst.read_text() == "dst"
    assert not tmp.exists()


def test_atomic_writer_failure(tmp_path):
    dst = tmp_path / "dst"
    with pytest.raises(Exception):
        with atomic_writer(dst) as tmp:
            tmp.write_text("dst")
            raise Exception("test")

    assert not dst.exists()
    assert not tmp.exists()


def test_atomic_writer_overwrite_symlink(tmp_path):
    target = tmp_path / "target"
    target.write_text("target")
    dst = tmp_path / "link"
    dst.symlink_to(target)

    with atomic_writer(dst) as tmp:
        tmp.write_text("dst")

    assert dst.read_text() == "dst"
    assert not dst.is_symlink()
    assert target.read_text() == "target"
    assert not tmp.exists()



from itertools import groupby
from operator import attrgetter

from jobrunner.lib.database import find_where
from jobrunner.models import Job


def calculate_workspace_state(workspace):
    """
    Return a list containing the most recent uncancelled job (if any) for each action in the workspace. We always
    ignore cancelled jobs when considering the historical state of the system. We also ignore jobs whose action is
    '__error__'; these are dummy jobs created only to help us communicate failure states back to the job-server (see
    create_or_update_jobs.create_failed_job()).
    """
    all_jobs = find_where(Job, workspace=workspace, cancelled=False)
    latest_jobs = []
    for action, jobs in group_by(all_jobs, attrgetter("action")):
        if action == "__error__":
            continue
        ordered_jobs = sorted(jobs, key=attrgetter("created_at"), reverse=True)
        latest_jobs.append(ordered_jobs[0])
    return latest_jobs


def group_by(iterable, key):
    return groupby(sorted(iterable, key=key), key=key)



"""
Script runs both jobrunner flows in a single process.
"""
import logging
import threading
import time

from jobrunner import config, record_stats, run, sync
from jobrunner.lib.log_utils import configure_logging


log = logging.getLogger(__name__)


def main():
    """Run the main run loop after starting the sync loop in a thread."""
    # extra space to align with other thread's "sync" label.
    threading.current_thread().name = "run "
    fmt = "{asctime} {threadName} {message} {tags}"
    configure_logging(fmt)

    try:
        log.info("jobrunner.service started")
        # daemon=True means this thread will be automatically join()ed when the
        # process exits
        thread = threading.Thread(target=sync_wrapper, daemon=True)
        thread.name = "sync"
        thread.start()
        # Stat the `record_stats` thread
        thread = threading.Thread(target=record_stats_wrapper, daemon=True)
        thread.name = "stat"
        thread.start()
        run.main()
    except KeyboardInterrupt:
        log.info("jobrunner.service stopped")


def sync_wrapper():
    """Wrap the sync call with logging context and an exception handler."""
    # avoid busy retries on hard failure
    sleep_after_error = config.POLL_INTERVAL * 5
    while True:
        try:
            sync.main()
        except sync.SyncAPIError as e:
            # Handle these separately as we don't want the full traceback here,
            # just the text of the error response
            log.error(e)
            time.sleep(sleep_after_error)
        except Exception:
            log.exception("Exception in sync thread")
            time.sleep(sleep_after_error)


def record_stats_wrapper():
    """Wrap the record_stats call with logging context and an exception handler."""
    while True:
        try:
            record_stats.main()
            # `main()` should loop forever, if it exits cleanly that means it
            # wasn't configured to run so we should exit the thread rather than
            # looping
            return
        except Exception:
            log.exception("Exception in record_stats thread")
            time.sleep(config.STATS_POLL_INTERVAL)


if __name__ == "__main__":
    main()



import sys


def main(output_file):
    """
    Write a dummy file to disk

    Databuilder doesn't currently provide many features for studies, so the
    options we have for actions in this full_project fixture are limited.  This
    "action" does the bare minimum just so we can have another action in the
    project.yaml.  Once databuilder supports more column types in the dummy
    data then we can remove this in favour of more meaningful actions such as
    those used in the cohortextractor version of the integration test.
    """
    with open(output_file, "w") as f:
        f.write("test file\n")


if __name__ == "__main__":
    main(*sys.argv[1:])



from urllib.parse import urlparse

from jobrunner.lib.git import commit_reachable_from_ref


class GithubValidationError(Exception):
    pass


def validate_repo_url(repo_url, allowed_gitub_orgs):
    parsed_url = urlparse(repo_url)
    if parsed_url.scheme != "https" or parsed_url.netloc != "github.com":
        raise GithubValidationError("Repository URLs must start https://github.com")
    path = parsed_url.path.strip("/").split("/")
    if not path or path[0] not in allowed_gitub_orgs:
        raise GithubValidationError(
            f"Repositories must belong to one of the following Github "
            f"organisations: {' '.join(allowed_gitub_orgs)}"
        )
    expected_url = f"https://github.com/{'/'.join(path[:2])}"
    if repo_url.rstrip("/") != expected_url or len(path) != 2:
        raise GithubValidationError(
            "Repository URL was not of the expected format: "
            "https://github.com/[organisation]/[project-name]"
        )


def validate_branch_and_commit(repo_url, commit, branch):
    """
    Due to the way Github works, anyone who can open a pull request against a
    repository can make a commit appear to be "in" that repository, even if
    they do not have write access to it.

    For example, someone created this PR against the Linux kernel:
    https://github.com/torvalds/linux/pull/437

    And even though this will never be merged, it still appears as a commit in
    that repo:
    https://github.com/torvalds/linux/commit/2793ae1df012c7c3f13ea5c0f0adb99017999c3b

    If we are enforcing that only code from certain organisations can be run
    then we need to check that any commits supplied have been made by someone
    with write access to the repository, which means we need to check they
    belong to a branch or tag in the repository.
    """
    if not branch:
        raise GithubValidationError("A branch name must be supplied")
    # A further wrinkle is that each PR gets an associated ref within the repo
    # of the form `pull/PR_NUMBER/head`. So we enforce that the branch name
    # must be a "plain vanilla" branch name with no slashes.
    if "/" in branch:
        raise GithubValidationError(f"Branch name must not contain slashes: {branch}")
    if not commit_reachable_from_ref(repo_url, commit, branch):
        raise GithubValidationError(
            f"Could not find commit on branch '{branch}': {commit}"
        )



import configparser
import os
import re
from multiprocessing import cpu_count
from pathlib import Path


class ConfigException(Exception):
    pass


def _is_valid_backend_name(name):
    return bool(re.match(r"^[A-Za-z0-9][A-Za-z0-9_\-]*[A-Za-z0-9]$", name))


default_work_dir = Path(__file__) / "../../workdir"

WORKDIR = Path(os.environ.get("WORKDIR", default_work_dir)).resolve()

TMP_DIR = WORKDIR / "temp"

GIT_REPO_DIR = WORKDIR / "repos"

DATABASE_FILE = WORKDIR / "db.sqlite"
DATABASE_SCHEMA_FILE = Path(__file__).parent / "schema.sql"

HIGH_PRIVACY_STORAGE_BASE = Path(
    os.environ.get("HIGH_PRIVACY_STORAGE_BASE", WORKDIR / "high_privacy")
)
MEDIUM_PRIVACY_STORAGE_BASE = Path(
    os.environ.get("MEDIUM_PRIVACY_STORAGE_BASE", WORKDIR / "medium_privacy")
)

HIGH_PRIVACY_WORKSPACES_DIR = HIGH_PRIVACY_STORAGE_BASE / "workspaces"
MEDIUM_PRIVACY_WORKSPACES_DIR = MEDIUM_PRIVACY_STORAGE_BASE / "workspaces"

HIGH_PRIVACY_ARCHIVE_DIR = os.environ.get(
    "HIGH_PRIVACY_ARCHIVE_DIR", HIGH_PRIVACY_STORAGE_BASE / "archives"
)

# valid archive formats
ARCHIVE_FORMATS = (".tar.gz", ".tar.zstd", ".tar.xz")


JOB_LOG_DIR = HIGH_PRIVACY_STORAGE_BASE / "logs"

JOB_SERVER_ENDPOINT = os.environ.get(
    "JOB_SERVER_ENDPOINT", "https://jobs.opensafely.org/api/v2/"
)
JOB_SERVER_TOKEN = os.environ.get("JOB_SERVER_TOKEN", "token")

PRIVATE_REPO_ACCESS_TOKEN = os.environ.get("PRIVATE_REPO_ACCESS_TOKEN", "")

POLL_INTERVAL = float(os.environ.get("POLL_INTERVAL", "5"))
JOB_LOOP_INTERVAL = float(os.environ.get("JOB_LOOP_INTERVAL", "1.0"))

BACKEND = os.environ.get("BACKEND", "expectations")
if not _is_valid_backend_name(BACKEND):
    raise RuntimeError(f"BACKEND not in valid format: '{BACKEND}'")

truthy = ("true", "1", "yes")

if os.environ.get("USING_DUMMY_DATA_BACKEND", "false").lower().strip() in truthy:
    USING_DUMMY_DATA_BACKEND = True
else:
    USING_DUMMY_DATA_BACKEND = BACKEND == "expectations"

ALLOWED_IMAGES = {
    "cohortextractor",
    "cohortextractor-v2",
    "databuilder",
    "stata-mp",
    "r",
    "jupyter",
    "python",
}

DOCKER_REGISTRY = "ghcr.io/opensafely-core"

DATABASE_URLS = {
    "full": os.environ.get("FULL_DATABASE_URL"),
    "slice": os.environ.get("SLICE_DATABASE_URL"),
    "dummy": os.environ.get("DUMMY_DATABASE_URL"),
}

TEMP_DATABASE_NAME = os.environ.get("TEMP_DATABASE_NAME")

EMIS_ORGANISATION_HASH = os.environ.get("EMIS_ORGANISATION_HASH")
PRESTO_TLS_KEY = PRESTO_TLS_CERT = None
PRESTO_TLS_CERT_PATH = os.environ.get("PRESTO_TLS_CERT_PATH")
PRESTO_TLS_KEY_PATH = os.environ.get("PRESTO_TLS_KEY_PATH")

if bool(PRESTO_TLS_KEY_PATH) != bool(PRESTO_TLS_CERT_PATH):
    raise ConfigException(
        "Both PRESTO_TLS_KEY_PATH and PRESTO_TLS_CERT_PATH must be defined if either are"
    )

if PRESTO_TLS_KEY_PATH:
    key_path = Path(PRESTO_TLS_KEY_PATH)
    if key_path.exists():
        PRESTO_TLS_KEY = key_path.read_text()
    else:
        raise ConfigException(
            f"PRESTO_TLS_KEY_PATH={key_path}, but file does not exist"
        )

if PRESTO_TLS_CERT_PATH:
    cert_path = Path(PRESTO_TLS_CERT_PATH)
    if cert_path.exists():
        PRESTO_TLS_CERT = cert_path.read_text()
    else:
        raise ConfigException(
            f"PRESTO_TLS_CERT_PATH={cert_path}, but file does not exist"
        )


MAX_WORKERS = int(os.environ.get("MAX_WORKERS") or max(cpu_count() - 1, 1))

# This is a crude mechanism for preventing a single large JobRequest with lots
# of associated Jobs from hogging all the resources. We want this configurable
# because it's useful to be able to disable this during tests and when running
# locally
RANDOMISE_JOB_ORDER = True

# Automatically delete containers and volumes after they have been used
CLEAN_UP_DOCKER_OBJECTS = True

# See `manage_jobs.ensure_overwritable` for more detail
ENABLE_PERMISSIONS_WORKAROUND = bool(os.environ.get("ENABLE_PERMISSIONS_WORKAROUND"))

STATA_LICENSE = os.environ.get("STATA_LICENSE")
STATA_LICENSE_REPO = os.environ.get(
    "STATA_LICENSE_REPO",
    "https://github.com/opensafely/server-instructions.git",
)


ACTIONS_GITHUB_ORG = "opensafely-actions"
ACTIONS_GITHUB_ORG_URL = f"https://github.com/{ACTIONS_GITHUB_ORG}"

ALLOWED_GITHUB_ORGS = (
    os.environ.get("ALLOWED_GITHUB_ORGS", "opensafely").strip().split(",")
)

# We hardcode this for now, as from a security perspective, we do not want it
# to be run time configurable. Though we do override this in `local_run.py` as
# we don't want to push traffic via the proxy when running locally.
GIT_PROXY_DOMAIN = "github-proxy.opensafely.org"


def parse_job_resource_weights(config_file):
    """
    Parse a simple ini file which looks like this:

        [some-workspace-name]
        my-ram-hungry-action = 4
        other-actions.* = 1.5

        [other-workspace-name]
        ...

    Any jobs in the specified workspace will have their action names matched
    against the regex patterns specified in the config file and will be
    assigned the weight of the first matching pattern. All other jobs are
    assigned a weight of 1.
    """
    config_file = Path(config_file)
    weights = {}
    if config_file.exists():
        config = configparser.ConfigParser()
        config.read_string(config_file.read_text(), source=str(config_file))
        for workspace in config.sections():
            weights[workspace] = {
                re.compile(pattern): float(weight)
                for (pattern, weight) in config.items(workspace)
            }
    return weights


JOB_RESOURCE_WEIGHTS = parse_job_resource_weights("job-resource-weights.ini")


STATS_DATABASE_FILE = os.environ.get("STATS_DATABASE_FILE")
if STATS_DATABASE_FILE:
    STATS_DATABASE_FILE = Path(STATS_DATABASE_FILE)

STATS_POLL_INTERVAL = float(os.environ.get("STATS_POLL_INTERVAL", "10"))


# feature flag to enable new API abstraction
EXECUTION_API = os.environ.get("EXECUTION_API", "true").lower() == "true"
EXECUTOR = os.environ.get("EXECUTOR", "jobrunner.executors.local:LocalDockerAPI")



import json
import sys
from contextlib import contextmanager

from jobrunner import config, models
from jobrunner.lib import database
from jobrunner.manage_jobs import MANIFEST_FILE, METADATA_DIR
from jobrunner.models import Job, State, isoformat_to_timestamp


def migrate_all(batch_size=10, log=True, dry_run=False, ignore_errors=False):
    _migrate(
        sorted(
            config.HIGH_PRIVACY_WORKSPACES_DIR.iterdir()
        ),  # sorted to make testing easier
        batch_size,
        write_medium_privacy_manifest=True,
        log=log,
        dry_run=dry_run,
        ignore_errors=ignore_errors,
    )


def migrate_one(
    workspace_dir,
    write_medium_privacy_manifest,
    batch_size=10,
    log=True,
    dry_run=False,
    ignore_errors=False,
):
    _migrate(
        [workspace_dir],
        batch_size,
        write_medium_privacy_manifest,
        log,
        dry_run,
        ignore_errors,
    )


def _migrate(
    workspace_dirs,
    batch_size,
    write_medium_privacy_manifest,
    log,
    dry_run,
    ignore_errors,
):
    count = 0

    for job in _jobs_from_workspaces(
        workspace_dirs, write_medium_privacy_manifest, log, dry_run, ignore_errors
    ):
        if count >= batch_size:
            _log(
                f"Reached batch size of {batch_size}. There are more jobs to be migrated.",
                log,
            )
            break

        if database.exists_where(Job, id=job.id):
            continue

        _insert_in_database(job, log, dry_run)
        count += 1

    if count == 0:
        _log("There were no jobs to migrate.", log)


def _jobs_from_workspaces(
    workspace_dirs, write_medium_privacy_manifest, log, dry_run, ignore_errors
):
    for workspace_dir in workspace_dirs:
        with _possibly_ignored_errors(log, ignore_errors, workspace=workspace_dir.name):
            yield from _jobs_from_workspace(
                workspace_dir,
                write_medium_privacy_manifest,
                log,
                dry_run,
                ignore_errors,
            )


def _jobs_from_workspace(
    workspace_dir, write_medium_privacy_manifest, log, dry_run, ignore_errors
):
    manifest_file = workspace_dir / METADATA_DIR / MANIFEST_FILE
    if not manifest_file.exists():
        return

    manifest = json.load(manifest_file.open())
    workspace_name = manifest.get("workspace", workspace_dir.name)
    repo = manifest.get("repo")
    all_files = manifest.get("files", {}).items()
    _log(f"Migrating workspace {workspace_name} in directory {workspace_dir}.", log)

    for action, action_details in manifest["actions"].items():
        with _possibly_ignored_errors(
            log, ignore_errors, workspace=workspace_name, action=action
        ):
            files = {
                file: file_details
                for file, file_details in all_files
                if file_details["created_by_action"] == action
            }
            yield _action_to_job(workspace_name, repo, files, action, action_details)

    _migrate_manifest_files(
        manifest_file, repo, workspace_name, write_medium_privacy_manifest, dry_run
    )


def _migrate_manifest_files(
    manifest, repo, workspace, write_medium_privacy_manifest, dry_run
):
    if dry_run:
        return

    if write_medium_privacy_manifest:
        _migrate_medium_privacy_manifest(repo, workspace)

    # We're done with this manifest. Move it out of the way so it doesn't cause confusion -- it's no longer being
    # updated, so soon it will be out of date.
    manifest.rename(manifest.with_name(f".deprecated.{MANIFEST_FILE}"))


def _migrate_medium_privacy_manifest(repo, workspace):
    # osrelease needs to be able to read the workspace name and repo URL from somewhere, in order to avoid the
    # person doing the release having to enter all the details. So we write this rump manifest just into the
    # medium privacy workspace. release-hatch is launched with this information already provided, so when osrelease
    # has been removed we can remove these files.
    manifest = (
        config.MEDIUM_PRIVACY_WORKSPACES_DIR / workspace / METADATA_DIR / MANIFEST_FILE
    )
    manifest.parent.mkdir(exist_ok=True, parents=True)
    manifest.write_text(json.dumps({"workspace": workspace, "repo": repo}))


def _action_to_job(workspace, repo, files, action, details):
    job_id = details["job_id"]
    if job_id == "unknown":
        # There are lots of historical jobs with their id recorded in the manifest as literally "unknown". The id is a
        # primary key, so we need unique values. But we don't need to join it to anything, so a random value is fine.
        job_id = models.random_id()

    return Job(
        id=job_id,
        workspace=workspace,
        repo_url=repo,
        action=action,
        state=_map_get(details, "state", State, None),
        commit=details.get("commit"),
        image_id=details.get("docker_image_id"),
        created_at=_map_get(details, "created_at", isoformat_to_timestamp, 0),
        completed_at=_map_get(details, "completed_at", isoformat_to_timestamp, 0),
        outputs={
            file: file_details["privacy_level"] for file, file_details in files.items()
        },
    )


def _insert_in_database(job, log, dry_run):
    if dry_run:
        log_prefix = "Dry run. Would have inserted"
    else:
        database.insert(job)
        log_prefix = "Inserted"
    _log(f"{log_prefix} Job(id={job.id}, action={job.action}).", log)


@contextmanager
def _possibly_ignored_errors(log, ignore_errors, **context):
    try:
        yield
    except Exception as e:
        if ignore_errors:
            _log(
                f"Ignoring {e.__class__.__name__}({e}) in {context}.",
                log,
            )
        else:
            raise


def _log(message, log):
    if log:
        print(message, file=sys.stderr)


def _map_get(mapping, key, func, default):
    try:
        return func(mapping[key])
    except (ValueError, KeyError):
        return default



import json

from jobrunner.lib.subprocess_utils import subprocess_run


def get_volume_and_container_sizes():
    response = subprocess_run(
        ["docker", "system", "df", "--verbose", "--format", "{{json .}}"],
        capture_output=True,
        check=True,
    )
    data = json.loads(response.stdout)
    volumes = {row["Name"]: _parse_size(row["Size"]) for row in data["Volumes"]}
    containers = {row["Names"]: _parse_size(row["Size"]) for row in data["Containers"]}
    return volumes, containers


def get_container_stats():
    response = subprocess_run(
        ["docker", "stats", "--no-stream", "--format", "{{json .}}"],
        capture_output=True,
        check=True,
    )
    data = [json.loads(line) for line in response.stdout.splitlines()]
    return {
        row["Name"]: {
            "cpu_percentage": float(row["CPUPerc"].rstrip("%")),
            "memory_used": _parse_size(row["MemUsage"].split()[0]),
        }
        for row in data
    }


CONVERSIONS = {
    "B": 1,
    "KB": 10**3,
    # Is this correct?
    "kB": 10**3,
    "KiB": 2**10,
    "MB": 10**6,
    "MiB": 2**20,
    "GB": 10**9,
    "GiB": 2**30,
    "TB": 10**12,
    "TiB": 2**40,
}


def _parse_size(size):
    units = size.lstrip("0123456789.-")
    value = float(size[: -len(units)])
    return int(value * CONVERSIONS[units])



from jobrunner.lib.docker import (
    MANAGEMENT_CONTAINER_IMAGE,
    DockerDiskSpaceError,
    docker,
)


__all__ = ["get_system_stats", "DockerDiskSpaceError"]

# Populated by calls to `register_command` below
COMMANDS = []
PARSERS = []


def get_system_stats():
    separator = "____"
    command = f" && echo {separator} && ".join(COMMANDS)
    response = docker(
        ["run", "--rm", MANAGEMENT_CONTAINER_IMAGE, "sh", "-c", command],
        capture_output=True,
        check=True,
    )
    output = response.stdout.decode("ascii", "ignore")
    stats = {}
    chunks = output.split(f"\n{separator}\n")
    for parser, chunk in zip(PARSERS, chunks):
        stats.update(parser(chunk))
    return stats


def register_command(command):
    def register_parser(fn):
        COMMANDS.append(command)
        PARSERS.append(fn)
        return fn

    return register_parser


# `-b`: output in bytes
@register_command("free -b")
def parse_output_from_free(output):
    # Example:
    #
    #               total        used        free      shared  buff/cache   available
    # Mem:    16640380928  6100041728   161775616  1044062208 10378563584  9152172032
    # Swap:   17040404480  3095134208 13945270272

    # Add a `type` header because the output doesn't include one
    output = "type " + output.strip()
    for row in _parse_table(output):
        if row["type"] == "Mem:":
            return {
                "total_memory": int(row["total"]),
                "available_memory": int(row["available"]),
            }


# `-k`: output in kilobytes
# `-P`: output in POSIX format
@register_command("df -kP /")
def parse_output_from_df(output):
    # Example:
    #
    # Filesystem           1024-blocks    Used Available Capacity Mounted on
    # overlay              967482320 639067280 279246760  70% /

    rows = _parse_table(output)
    assert len(rows) == 1
    data = rows[0]
    assert data["Mounted"] == "/"
    return {
        "total_disk_space": int(data["1024-blocks"]) * 1024,
        "available_disk_space": int(data["Available"]) * 1024,
    }


# `-P ALL`: show results for all processors
# `1 1`: produce 1 report, after an interval of 1 second
@register_command("mpstat -P ALL 1 1")
def parse_output_from_mpstat(output):
    # Example:
    # Linux 5.4.0-66-generic (c30c70806105)	03/09/21	_x86_64_	(4 CPU)
    #
    # 11:02:35     CPU    %usr   %nice    %sys %iowait    %irq   %soft  %steal  %guest   %idle
    # 11:02:36     all   22.31    0.25    7.27    1.50    0.00    1.75    0.00    0.00   66.92
    # 11:02:36       0   14.29    0.00    3.06    1.02    0.00    1.02    0.00    0.00   80.61
    # 11:02:36       1   20.62    0.00    7.22    5.15    0.00    0.00    0.00    0.00   67.01
    # 11:02:36       2   16.49    0.00    5.15    0.00    0.00    6.19    0.00    0.00   72.16
    # 11:02:36       3   36.27    0.00   12.75    0.00    0.00    0.00    0.00    0.00   50.98
    #
    # Average:     CPU    %usr   %nice    %sys %iowait    %irq   %soft  %steal  %guest   %idle
    # Average:     all   22.31    0.25    7.27    1.50    0.00    1.75    0.00    0.00   66.92
    # Average:       0   14.29    0.00    3.06    1.02    0.00    1.02    0.00    0.00   80.61
    # Average:       1   20.62    0.00    7.22    5.15    0.00    0.00    0.00    0.00   67.01
    # Average:       2   16.49    0.00    5.15    0.00    0.00    6.19    0.00    0.00   72.16
    # Average:       3   36.27    0.00   12.75    0.00    0.00    0.00    0.00    0.00   50.98
    prefix = "Average:"
    output = "\n".join(
        line[len(prefix) :] for line in output.splitlines() if line.startswith(prefix)
    )
    rows = _parse_table(output)
    rows = [
        {
            key: float(value) if key.startswith("%") else value
            for (key, value) in row.items()
        }
        for row in rows
    ]
    return {"mpstat": rows}


def _parse_table(table_str):
    table = [line.split() for line in table_str.strip().splitlines()]
    header, rows = table[0], table[1:]
    return [dict(zip(header, row)) for row in rows]



import pytest

from jobrunner.lib.system_stats import get_system_stats


@pytest.mark.needs_docker
def test_get_system_stats():
    stats = get_system_stats()
    assert {
        "total_disk_space",
        "available_disk_space",
        "total_memory",
        "available_memory",
    }.issubset(stats.keys())



"""
Ops utility for killing jobs and cleaning up containers and volumes
"""
import argparse

from jobrunner.lib.database import find_where
from jobrunner.manage_jobs import container_name, docker, volume_name
from jobrunner.models import Job, State
from jobrunner.run import mark_job_as_failed


def main(partial_job_ids, cleanup=False):
    jobs = get_jobs(partial_job_ids)
    for job in jobs:
        # If the job has been previously killed we don't want to overwrite the
        # timestamps here
        if job.state in (State.PENDING, State.RUNNING):
            mark_job_as_failed(job, "Killed by admin")
        # All these docker commands are idempotent
        docker.kill(container_name(job))
        if cleanup:
            docker.delete_container(container_name(job))
            docker.delete_volume(volume_name(job))


def get_jobs(partial_job_ids):
    jobs = []
    need_confirmation = False
    for partial_job_id in partial_job_ids:
        matches = find_where(Job, id__like=f"%{partial_job_id}%")
        if len(matches) == 0:
            raise RuntimeError(f"No jobs found matching '{partial_job_id}'")
        elif len(matches) > 1:
            print(f"Multiple jobs found matching '{partial_job_id}':")
            for i, job in enumerate(matches, start=1):
                print(f"  {i}: {job.slug}")
            print()
            index = int(input("Enter number: "))
            assert 0 < index <= len(matches)
            jobs.append(matches[index - 1])
        else:
            need_confirmation = True
            jobs.append(matches[0])
    if need_confirmation:
        print("About to kill jobs:")
        for job in jobs:
            print(f"  {job.slug}")
        confirm = input("\nEnter to continue, Ctrl-C to quit ")
        assert confirm == ""
    return jobs


def run():
    parser = argparse.ArgumentParser(description=__doc__.partition("\n\n")[0])
    parser.add_argument(
        "--cleanup",
        action="store_true",
        help="Delete any associated containers and volumes",
    )
    parser.add_argument(
        "partial_job_ids", nargs="+", help="ID of the job (or substring of the ID)"
    )
    args = parser.parse_args()
    main(**vars(args))


if __name__ == "__main__":
    run()



"""
This sets up some basic logging configuration (log to stderr with a default log
level of INFO) with some job-runner specific tweaks. In particular it supports
automatically including the currently executing Job or JobRequest in the log
output. It also includes the stderr output from any failed attempts to shell
out to external processes.
"""
import contextlib
import logging
import logging.handlers
import os
import subprocess
import sys
import threading
import time


DEFAULT_FORMAT = "{asctime} {message} {tags}"


def configure_logging(
    fmt=DEFAULT_FORMAT,
    stream=None,
    extra_filter=None,
):
    formatter = JobRunnerFormatter(fmt, style="{")
    handler = logging.StreamHandler(stream=stream)
    handler.setFormatter(formatter)
    if extra_filter:
        handler.addFilter(extra_filter)
    handler.addFilter(formatting_filter)

    log_level = os.environ.get("LOGLEVEL", "INFO")
    handlers = [handler]

    # Support a separate log file at level DEBUG, while leaving the default
    # logs untouched. DEBUG logging can be extremely noisy and so we want a way
    # to capture these that doesn't pollute the primary logs.
    debug_log_file = os.environ.get("DEBUG_LOG_FILE")
    if debug_log_file:
        debug_handler = logging.handlers.TimedRotatingFileHandler(
            debug_log_file,
            encoding="utf-8",
            delay=True,
            # Rotate daily, keeping 14 days of backups
            when="D",
            interval=1,
            backupCount=14,
            utc=True,
        )
        debug_handler.setFormatter(formatter)
        debug_handler.addFilter(formatting_filter)
        debug_handler.setLevel("DEBUG")
        handlers.append(debug_handler)
        # Set the default handler to the originally specified log level and
        # then increase the base log level to DEBUG
        handler.setLevel(log_level)
        log_level = "DEBUG"

    logging.basicConfig(level=log_level, handlers=handlers)

    if debug_log_file:
        logging.getLogger(__name__).info(f"Writing DEBUG logs to '{debug_log_file}'")

    # We attach a custom handler for uncaught exceptions to display error
    # output from failed subprocesses
    sys.excepthook = show_subprocess_stderr


class JobRunnerFormatter(logging.Formatter):

    converter = time.gmtime  # utc rather than local
    default_msec_format = "%s.%03dZ"  # s/,/. and append Z

    def formatException(self, exc_info):
        """
        We frequently shell out to docker and git and it's very useful to have
        the stderr output in the logs rather than just "this process failed"
        """
        message = super().formatException(exc_info)
        value = exc_info[1]
        if isinstance(value, subprocess.CalledProcessError):
            stderr = value.stderr
            if stderr:
                if isinstance(stderr, bytes):
                    stderr = stderr.decode("utf-8", "ignore")
                message = f"{message}\n\nstderr:\n\n{stderr}"
        return message


def formatting_filter(record):
    """Ensure various record attribute are always available for formatting."""

    ctx = set_log_context.current_context
    job = getattr(record, "job", None) or ctx.get("job")
    req = getattr(record, "job_request", None) or ctx.get("job_request")

    status_code = getattr(record, "status_code", None)
    if job and not status_code:
        status_code = job.status_code

    tags = {}

    if status_code:
        tags["status"] = status_code
    if job:
        tags["project"] = job.project
        tags["action"] = job.action
        tags["id"] = job.id
    if req:
        tags["req"] = req.id

    record.tags = " ".join(f"{k}={v}" for k, v in tags.items())

    # The `action` attribute is only used by format string in "local_run" mode
    # but we make sure it's always available
    record.action = f"{job.action}: " if job else ""

    return True


class SetLogContext(threading.local):
    """
    A context manager which allows setting `extra` values on all logging calls
    which occur anywhere in its context e.g.

        set_log_context = SetLogContext()
        with set_log_context(foo="bar"):
            log.info("hello world")

    Will result in the equivalent log message to:

        log.info("hello word", extra={"foo": "bar"})

    Uses threading.local for thread safety.
    """

    def __init__(self):
        self.current_context = {}
        self.context_stack = []

    @contextlib.contextmanager
    def __call__(self, **kwargs):
        """
        Create a new logging context with the supplied keyword arguments (in
        addition to any inherited from the current context)
        """
        self.context_stack.append(self.current_context)
        self.current_context = dict(self.current_context, **kwargs)
        try:
            yield
        finally:
            self.current_context = self.context_stack.pop()


def show_subprocess_stderr(typ, value, traceback):
    """
    This applies the same CalledProcessError formatting as in `JobRunnerFormatter`
    above but to uncaught exceptions
    """
    sys.__excepthook__(typ, value, traceback)
    if isinstance(value, subprocess.CalledProcessError):
        stderr = value.stderr
        if stderr:
            if isinstance(stderr, bytes):
                stderr = stderr.decode("utf-8", "ignore")
            print("\nstderr:\n", file=sys.stderr)
            print(stderr, file=sys.stderr)


set_log_context = SetLogContext()



from ruamel.yaml import YAML, error


class YAMLError(Exception):
    pass


def parse_yaml(file_contents, name="yaml file"):
    """
    Parse a YAML file supplied as bytes into a dictionary

    Args:
        file_contents: file contents as bytes
        name: optional name of the file for producing more readable error
            messages

    Returns:
        parsed contents as a dictionary

    Raises:
        YAMLError
    """
    try:
        # We're using the pure-Python version here as we don't care about speed
        # and this gives better error messages (and consistent behaviour
        # cross-platform)
        return YAML(typ="safe", pure=True).load(file_contents)
    # ruamel doesn't have a nice exception hierarchy so we have to catch these
    # four separate base classes
    except (
        error.YAMLError,
        error.YAMLStreamError,
        error.YAMLWarning,
        error.YAMLFutureWarning,
    ) as exc:
        exc = make_yaml_error_more_helpful(exc, name)
        raise YAMLError(f"{type(exc).__name__} {exc}")


def make_yaml_error_more_helpful(exc, name):
    """
    ruamel produces quite helpful error messages but they refer to the file as
    `<byte_string>` (which will be confusing for users) and they also include
    notes and warnings to developers about API changes. This function attempts
    to fix these issues, but just returns the exception unchanged if anything
    goes wrong.
    """
    try:
        try:
            exc.context_mark.name = name
        except AttributeError:
            pass
        try:
            exc.problem_mark.name = name
        except AttributeError:
            pass
        exc.note = ""
        exc.warn = ""
    except Exception:
        pass
    return exc



import os


def list_dir_with_ignore_patterns(directory, ignore_patterns):
    """
    Given a directory and list of glob patterns, return all paths in that
    directory which don't match any of the glob patterns.

    Note that this function won't descend further than it needs to so, for
    instance, if there is a top level directory "foo" and no ignore pattern
    begins with "foo/" then it will return "foo" without iterating any of the
    files or sub-directories within it.
    """
    matches = []
    for pattern in ignore_patterns:
        matches.extend(directory.glob(pattern))
    match_tree = {}
    for match in matches:
        relative_match = match.relative_to(directory)
        tree = match_tree
        for segment in str(relative_match).split(os.sep):
            if segment not in tree:
                tree[segment] = {}
            tree = tree[segment]
    return [path.relative_to(directory) for path in _iter_dir(directory, match_tree)]


def _iter_dir(directory, match_tree):
    for item in directory.iterdir():
        subtree = match_tree.get(item.name)
        # No match: yield this file
        if subtree is None:
            yield item
        # Match is a leaf node: ignore this specific file
        elif not subtree:
            pass
        # Otherwise filter this subdirectory using the subtree
        else:
            yield from _iter_dir(item, subtree)



import json
import logging
import subprocess
from pathlib import Path

from jobrunner import config
from jobrunner.job_executor import (
    ExecutorAPI,
    ExecutorState,
    JobDefinition,
    JobResults,
    JobStatus,
    Privacy,
)
from jobrunner.lib import docker
from jobrunner.lib.string_utils import tabulate

# ideally, these should be moved into this module when the old implementation
# is removed
from jobrunner.manage_jobs import (
    METADATA_DIR,
    TIMESTAMP_REFERENCE_FILE,
    cleanup_job,
    container_name,
    copy_file,
    copy_git_commit_to_volume,
    copy_local_workspace_to_volume,
    ensure_overwritable,
    get_container_metadata,
    get_high_privacy_workspace,
    get_log_dir,
    get_medium_privacy_workspace,
    volume_name,
    write_manifest_file,
)


# cache of result objects
RESULTS = {}
LABEL = "jobrunner-local"

log = logging.getLogger(__name__)


class LocalDockerError(Exception):
    pass


def get_job_labels(job: JobDefinition):
    """Useful metadata to label docker objects with."""
    return {
        "workspace": job.workspace,
        "action": job.action,
    }


def workspace_is_archived(workspace):
    archive_dir = config.HIGH_PRIVACY_ARCHIVE_DIR
    for ext in config.ARCHIVE_FORMATS:
        path = (archive_dir / workspace).with_suffix(ext)
        if path.exists():
            return True
    return False


class LocalDockerAPI(ExecutorAPI):
    """ExecutorAPI implementation using local docker service."""

    def prepare(self, job):
        current = self.get_status(job)
        if current.state != ExecutorState.UNKNOWN:
            return current

        # Check the workspace is not archived
        workspace_dir = get_high_privacy_workspace(job.workspace)
        if not workspace_dir.exists():
            if workspace_is_archived(job.workspace):
                return JobStatus(
                    ExecutorState.ERROR,
                    f"Workspace {job.workspace} has been archived. Contact the OpenSAFELY tech team to resolve",
                )

        # Check the image exists locally and error if not. Newer versions of
        # docker-cli support `--pull=never` as an argument to `docker run` which
        # would make this simpler, but it looks like it will be a while before this
        # makes it to Docker for Windows:
        # https://github.com/docker/cli/pull/1498
        if not docker.image_exists_locally(job.image):
            log.info(f"Image not found, may need to run: docker pull {job.image}")
            return JobStatus(
                ExecutorState.ERROR,
                f"Docker image {job.image} is not currently available",
            )

        try:
            prepare_job(job)
        except docker.DockerDiskSpaceError as e:
            log.exception(str(e))
            return JobStatus(
                ExecutorState.ERROR, "Out of disk space, please try again later"
            )

        # technically, we're acutally PREPARED, as we did in synchronously, but
        # the loop code is expecting PREPARING, so return that. The next time
        # around the loop, it will pick up that it is PREPARED, and move on.
        return JobStatus(ExecutorState.PREPARING)

    def execute(self, job):
        current = self.get_status(job)
        if current.state != ExecutorState.PREPARED:
            return current

        try:
            docker.run(
                container_name(job),
                [job.image] + job.args,
                volume=(volume_name(job), "/workspace"),
                env=job.env,
                allow_network_access=job.allow_database_access,
                label=LABEL,
                labels=get_job_labels(job),
            )
        except Exception as exc:
            return JobStatus(
                ExecutorState.ERROR, f"Failed to start docker container: {exc}"
            )

        return JobStatus(ExecutorState.EXECUTING)

    def finalize(self, job):

        current = self.get_status(job)
        if current.state != ExecutorState.EXECUTED:
            return current

        try:
            finalize_job(job)
        except LocalDockerError as exc:
            return JobStatus(ExecutorState.ERROR, f"failed to finalize job: {exc}")

        return JobStatus(ExecutorState.FINALIZING)

    def terminate(self, job):
        docker.kill(container_name(job))
        return JobStatus(ExecutorState.ERROR, "terminated by api")

    def cleanup(self, job):
        cleanup_job(job)
        RESULTS.pop(job.id, None)
        return JobStatus(ExecutorState.UNKNOWN)

    def get_status(self, job):
        name = container_name(job)
        job_running = docker.container_inspect(
            name, "State.Running", none_if_not_exists=True
        )

        if job_running is None:
            # no container for this job found
            volume = volume_name(job)
            if docker.volume_exists(volume):
                return JobStatus(ExecutorState.PREPARED)
            else:
                return JobStatus(ExecutorState.UNKNOWN)

        elif job_running:
            return JobStatus(ExecutorState.EXECUTING)
        elif job.id in RESULTS:
            return JobStatus(ExecutorState.FINALIZED)
        else:  # container present but not running, i.e. finished
            return JobStatus(ExecutorState.EXECUTED)

    def get_results(self, job):
        if job.id not in RESULTS:
            return JobStatus(ExecutorState.ERROR, "job has not been finalized")

        return RESULTS[job.id]

    def delete_files(self, workspace, privacy, files):
        if privacy == Privacy.HIGH:
            root = get_high_privacy_workspace(workspace)
        elif privacy == Privacy.MEDIUM:
            root = get_medium_privacy_workspace(workspace)
        else:
            raise Exception(f"unknown privacy of {privacy}")

        errors = []
        for name in files:
            path = root / name
            try:
                path.unlink(missing_ok=True)
            except Exception:
                log.exception(f"Could not delete {path}")
                errors.append(name)

        return errors


def prepare_job(job):
    """Creates a volume and populates it with the repo and input files."""
    workspace_dir = get_high_privacy_workspace(job.workspace)

    volume = volume_name(job)
    docker.create_volume(volume, get_job_labels(job))

    # `docker cp` can't create parent directories for us so we make sure all
    # these directories get created when we copy in the code
    extra_dirs = set(Path(filename).parent for filename in job.inputs)

    try:
        if job.study.git_repo_url and job.study.commit:
            copy_git_commit_to_volume(
                volume, job.study.git_repo_url, job.study.commit, extra_dirs
            )
        else:
            # We only encounter jobs without a repo or commit when using the
            # "local_run" command to execute uncommitted local code
            copy_local_workspace_to_volume(volume, workspace_dir, extra_dirs)
    except subprocess.CalledProcessError:
        raise LocalDockerError(
            f"Could not checkout commit {job.study.commit} from {job.study.git_repo_url}"
        )

    for filename in job.inputs:
        log.info(f"Copying input file: {filename}")
        if not (workspace_dir / filename).exists():
            raise LocalDockerError(
                f"The file {filename} doesn't exist in workspace {job.workspace} as requested for job {job.id}"
            )
        docker.copy_to_volume(volume, workspace_dir / filename, filename)

    # Hack: see `get_unmatched_outputs`. For some reason this requires a
    # non-empty file so copying `os.devnull` didn't work.
    some_non_empty_file = Path(__file__)
    docker.copy_to_volume(volume, some_non_empty_file, TIMESTAMP_REFERENCE_FILE)
    return volume


def finalize_job(job):
    container_metadata = get_container_metadata(job)
    outputs, unmatched_patterns = find_matching_outputs(job)
    exit_code = container_metadata["State"]["ExitCode"]
    message = None
    if exit_code == 137:
        # 137 = 128+9, which means was killed by signal 9, SIGKILL
        # This usually happens because of OOM killer, or else manually
        message = "likely means it ran out of memory"
    results = JobResults(
        outputs=outputs,
        unmatched_patterns=unmatched_patterns,
        exit_code=container_metadata["State"]["ExitCode"],
        image_id=container_metadata["Image"],
        message=message,
    )
    persist_outputs(job, results.outputs, container_metadata)
    RESULTS[job.id] = results


def persist_outputs(job, outputs, container_metadata):
    """Copy logs and generated outputs to persistant storage."""
    # job_metadata is a big dict capturing everything we know about the state
    # of the job
    job_metadata = dict()
    job_metadata["id"] = job.id
    job_metadata["docker_image_id"] = container_metadata["Image"]
    job_metadata["exit_code"] = container_metadata["State"]["ExitCode"]
    job_metadata["container_metadata"] = container_metadata
    job_metadata["outputs"] = outputs
    job_metadata["commit"] = job.study.commit

    # Dump useful info in log directory
    log_dir = get_log_dir(job)
    ensure_overwritable(log_dir / "logs.txt", log_dir / "metadata.json")
    write_log_file(job, job_metadata, log_dir / "logs.txt")
    with open(log_dir / "metadata.json", "w") as f:
        json.dump(job_metadata, f, indent=2)

    # Copy logs to workspace
    workspace_dir = get_high_privacy_workspace(job.workspace)
    metadata_log_file = workspace_dir / METADATA_DIR / f"{job.action}.log"
    copy_file(log_dir / "logs.txt", metadata_log_file)
    log.info(f"Logs written to: {metadata_log_file}")

    # Extract outputs to workspace
    ensure_overwritable(*[workspace_dir / f for f in outputs.keys()])
    volume = volume_name(job)
    for filename in outputs.keys():
        log.info(f"Extracting output file: {filename}")
        docker.copy_from_volume(volume, filename, workspace_dir / filename)

    # Copy out logs and medium privacy files
    medium_privacy_dir = get_medium_privacy_workspace(job.workspace)
    if medium_privacy_dir:
        copy_file(
            workspace_dir / METADATA_DIR / f"{job.action}.log",
            medium_privacy_dir / METADATA_DIR / f"{job.action}.log",
        )
        for filename, privacy_level in outputs.items():
            if privacy_level == "moderately_sensitive":
                copy_file(workspace_dir / filename, medium_privacy_dir / filename)

        # this can be removed once osrelease is dead
        write_manifest_file(
            medium_privacy_dir,
            {"repo": job.study.git_repo_url, "workspace": job.workspace},
        )


def find_matching_outputs(job):
    """
    Returns a dict mapping output filenames to their privacy level, plus a list
    of any patterns that had no matches at all
    """
    all_matches = docker.glob_volume_files(volume_name(job), job.output_spec.keys())
    unmatched_patterns = []
    outputs = {}
    for pattern, privacy_level in job.output_spec.items():
        filenames = all_matches[pattern]
        if not filenames:
            unmatched_patterns.append(pattern)
        for filename in filenames:
            outputs[filename] = privacy_level
    return outputs, unmatched_patterns


def write_log_file(job, job_metadata, filename):
    """
    This dumps the (timestamped) Docker logs for a job to disk, followed by
    some useful metadata about the job and its outputs
    """
    filename.parent.mkdir(parents=True, exist_ok=True)
    docker.write_logs_to_file(container_name(job), filename)
    outputs = sorted(job_metadata["outputs"].items())
    with open(filename, "a") as f:
        f.write("\n\n")
        for key in KEYS_TO_LOG:
            if not job_metadata[key]:
                continue
            f.write(f"{key}: {job_metadata[key]}\n")
        f.write("\noutputs:\n")
        f.write(tabulate(outputs, separator="  - ", indent=2, empty="(no outputs)"))


# Keys of fields to log in manifest.json and log file
KEYS_TO_LOG = [
    "id",
    "commit",
    "docker_image_id",
    "exit_code",
]



"""
Ops utility for retrying `finalise_job` on a job which hit an error during
completion

This only applies to (and only works for) jobs which failed with an
Internal Error during the "finalise" step where we copy output files,
dump logs etc. When this happens we don't automatically clean up the
job's container and volume giving us an opportunity to fix the bug and
try again without having to re-run the entire job.

To do this we simply put the job back into the RUNNING state and let the
jobrunner pick it up again. We also need to update the job-server when we do
this so that it puts the job back in an "active" state and continues to ask for
updates on it.
"""
import argparse
import time

from jobrunner.lib.database import find_where, update
from jobrunner.manage_jobs import container_name, docker
from jobrunner.models import Job, State
from jobrunner.sync import api_post, job_to_remote_format


def main(partial_job_id):
    job = get_job(partial_job_id)
    if not docker.container_exists(container_name(job)):
        raise RuntimeError("Cannot reset job, associated container does not exist")
    job.state = State.RUNNING
    job.status_message = "Re-attempting to extract outputs"
    job.status_code = None
    job.completed_at = None
    job.updated_at = int(time.time())
    print("\nUpdating job in database:")
    print(job)
    update(job)
    print("\nPOSTing update to job-server")
    api_post("jobs", json=[job_to_remote_format(job)])
    print("\nDone")


def get_job(partial_job_id):
    matches = find_where(Job, id__like=f"%{partial_job_id}%")
    if len(matches) == 0:
        raise RuntimeError("No matching jobs found")
    elif len(matches) > 1:
        print("Multiple matching jobs found:")
        for i, job in enumerate(matches, start=1):
            print(f"  {i}: {job.slug}")
        print()
        index = int(input("Enter number: "))
        assert 0 < index <= len(matches)
        job = matches[index - 1]
    else:
        job = matches[0]
        print(f"About to reset job:\n  {job.slug}\n")
        confirm = input("Enter to continue, Ctrl-C to quit ")
        assert confirm == ""
    return job


def run():
    parser = argparse.ArgumentParser(description=__doc__.partition("\n\n")[0])
    parser.add_argument("partial_job_id", help="ID of the job (or substring of the ID)")
    args = parser.parse_args()
    main(**vars(args))


if __name__ == "__main__":
    run()



"""
Ops utility for killing all running jobs and reseting them to PENDING so they will be
automatically re-run after a reboot.
"""
import argparse

from jobrunner.lib.database import find_where, update_where
from jobrunner.manage_jobs import container_name, docker, volume_name
from jobrunner.models import Job, State


def main():
    print(
        "== DANGER ZONE ==\n"
        "\n"
        "This will kill all running jobs and reset them to the PENDING state, ready\n"
        "to be restarted following a reboot.\n"
        "\n"
        "It should only be run when the job-runner service has been stopped."
        "\n"
    )
    confirm = input("Are you sure you want to continue? (y/N)")
    assert confirm.strip().lower() == "y"
    # Reset all running jobs to pending
    update_where(Job, {"state": State.PENDING, "started_at": None}, state=State.RUNNING)
    # Make sure all containers and volumes are removed ready to freshly restart the jobs
    # after the reboot
    for job in find_where(Job, state=State.PENDING):
        docker.kill(container_name(job))
        docker.delete_container(container_name(job))
        docker.delete_volume(volume_name(job))


def run():
    parser = argparse.ArgumentParser(description=__doc__.partition("\n\n")[0])
    args = parser.parse_args()
    main(**vars(args))


if __name__ == "__main__":
    run()



"""
Script which polls the database for active (i.e. non-terminated) jobs, takes
the appropriate action for each job depending on its current state, and then
updates its state as appropriate.
"""
import datetime
import logging
import random
import shlex
import sys
import time
from typing import Optional

from jobrunner import config
from jobrunner.executors import get_executor_api
from jobrunner.job_executor import (
    ExecutorAPI,
    ExecutorState,
    JobDefinition,
    Privacy,
    Study,
)
from jobrunner.lib.database import find_where, select_values, update
from jobrunner.lib.log_utils import configure_logging, set_log_context
from jobrunner.manage_jobs import (
    BrokenContainerError,
    JobError,
    cleanup_job,
    finalise_job,
    job_still_running,
    kill_job,
    list_outputs_from_action,
    start_job,
)
from jobrunner.models import Job, State, StatusCode
from jobrunner.project import requires_db_access


log = logging.getLogger(__name__)


class InvalidTransition(Exception):
    pass


def main(exit_callback=lambda _: False):
    log.info("jobrunner.run loop started")

    api = None
    if config.EXECUTION_API:
        log.info("using new EXECUTION_API")
        api = get_executor_api()

    while True:
        active_jobs = handle_jobs(api)

        if exit_callback(active_jobs):
            break
        time.sleep(config.JOB_LOOP_INTERVAL)


def handle_jobs(api: Optional[ExecutorAPI]):
    log.debug("Querying database for active jobs")
    active_jobs = find_where(Job, state__in=[State.PENDING, State.RUNNING])
    log.debug("Done query")
    # Randomising the job order is a crude but effective way to ensure that a
    # single large job request doesn't hog all the workers. We make this
    # optional as, when running locally, having jobs run in a predictable order
    # is preferable
    if config.RANDOMISE_JOB_ORDER:
        random.shuffle(active_jobs)

    for job in active_jobs:
        # `set_log_context` ensures that all log messages triggered anywhere
        # further down the stack will have `job` set on them
        with set_log_context(job=job):
            if api:
                handle_active_job_api(job, api)
            else:
                # old way
                if job.state == State.PENDING:
                    handle_pending_job(job)
                elif job.state == State.RUNNING:
                    handle_running_job(job)

    return active_jobs


def handle_pending_job(job):
    if job.cancelled:
        # Mark the job as running and then immediately invoke
        # `handle_running_job` to deal with the cancellation. This slightly
        # counterintuitive approach allows us to keep a simple, consistent set
        # of state transitions and to consolidate all the kill/cleanup code
        # together. It also means that there aren't edge cases where we could
        # lose track of jobs completely after losing database state
        mark_job_as_running(job)
        handle_running_job(job)
        return

    awaited_states = get_states_of_awaited_jobs(job)
    if State.FAILED in awaited_states:
        mark_job_as_failed(
            job, "Not starting as dependency failed", code=StatusCode.DEPENDENCY_FAILED
        )
    elif any(state != State.SUCCEEDED for state in awaited_states):
        set_message(
            job, "Waiting on dependencies", code=StatusCode.WAITING_ON_DEPENDENCIES
        )
    else:
        not_started_reason = get_reason_job_not_started(job)
        if not_started_reason:
            set_message(job, not_started_reason, code=StatusCode.WAITING_ON_WORKERS)
        else:
            try:
                set_message(job, "Preparing")
                start_job(job)
            except JobError as exception:
                mark_job_as_failed(job, exception)
                # See the `raise` in manage_jobs which explains why we can't
                # cleanup on this specific error
                if not isinstance(exception, BrokenContainerError):
                    cleanup_job(job)
            except Exception:
                mark_job_as_failed(job, "Internal error when starting job")
                cleanup_job(job)
                raise
            else:
                mark_job_as_running(job)


def handle_running_job(job):
    if job.cancelled:
        log.info("Cancellation requested, killing job")
        kill_job(job)

    log.debug("Checking job running state")
    is_running = job_still_running(job)
    log.debug("Check done")
    if is_running:
        set_message(job, "Running")
    else:
        try:
            set_message(job, "Finished, checking status and extracting outputs")
            job = finalise_job(job)
            # We expect the job to be transitioned into its final state at this
            # point
            assert job.state in [State.SUCCEEDED, State.FAILED]
        except JobError as exception:
            mark_job_as_failed(job, exception)
            # Question: do we want to clean up failed jobs? Given that we now
            # tag all job-runner volumes and containers with a specific label
            # we could leave them around for debugging purposes and have a
            # cronjob which cleans them up a few days after they've stopped.
            cleanup_job(job)
        except Exception:
            mark_job_as_failed(job, "Internal error when finalising job")
            # We deliberately don't clean up after an internal error so we have
            # some change of debugging. It's also possible, after fixing the
            # error, to manually flip the state of the job back to "running" in
            # the database and the code will then be able to finalise it
            # correctly without having to re-run the job.
            raise
        else:
            mark_job_as_completed(job)
            cleanup_job(job)


# we do not control the tranisition from these states, the executor does
STABLE_STATES = [
    ExecutorState.PREPARING,
    ExecutorState.EXECUTING,
    ExecutorState.FINALIZING,
]


def handle_active_job_api(job, api):
    try:
        handle_job_api(job, api)
    except Exception:
        mark_job_as_failed(job, "Internal error")
        # Do not clean up, as we may want to debug
        #
        # Raising will kill the main loop, by design. The service manager
        # will restart, and this job will be ignored when it does, as
        # it has failed. If we have an internal error, a full restart
        # might recover better.
        raise


def handle_job_api(job, api):
    """Handle an active job.

    This contains the main state machine logic for a job. For the most part,
    state transitions follow the same logic, which is abstracted. Some
    transitions require special logic, mainly the initial and final states, as
    well as supporting cancellation.
    """
    assert job.state in (State.PENDING, State.RUNNING)
    definition = job_to_job_definition(job)

    if job.cancelled:
        # cancelled is driven by user request, so is handled explicitly first
        # regardless of executor state.
        api.terminate(definition)
        mark_job_as_failed(job, "Cancelled by user", StatusCode.CANCELLED_BY_USER)
        api.cleanup(definition)
        return

    initial_status = api.get_status(definition)

    # handle the simple no change needed states.
    if initial_status.state in STABLE_STATES:
        if job.state == State.PENDING:
            log.warning(
                "state ereror: got {initial_status.state} for a job we thought was PENDING"
            )
        # no action needed, simply update job message and timestamp
        message = initial_status.state.value.title()
        set_message(job, message)
        return

    # ok, handle the state transitions that are our responsibility
    if initial_status.state == ExecutorState.UNKNOWN:
        # a new job
        if job.state == State.RUNNING:
            log.warning(
                "state error: got UNKNOWN state for a job we thought was RUNNING"
            )

        # check dependencies
        awaited_states = get_states_of_awaited_jobs(job)
        if State.FAILED in awaited_states:
            mark_job_as_failed(
                job,
                "Not starting as dependency failed",
                code=StatusCode.DEPENDENCY_FAILED,
            )
            return

        if any(state != State.SUCCEEDED for state in awaited_states):
            set_message(
                job,
                "Waiting on dependencies",
                code=StatusCode.WAITING_ON_DEPENDENCIES,
            )
            return

        expected_state = ExecutorState.PREPARING
        new_status = api.prepare(definition)

    elif initial_status.state == ExecutorState.PREPARED:
        expected_state = ExecutorState.EXECUTING
        new_status = api.execute(definition)

    elif initial_status.state == ExecutorState.EXECUTED:
        expected_state = ExecutorState.FINALIZING
        new_status = api.finalize(definition)

    elif initial_status.state == ExecutorState.FINALIZED:
        # final state - we have finished!
        results = api.get_results(definition)
        save_results(job, results)
        obsolete = get_obsolete_files(definition, results.outputs)
        if obsolete:
            errors = api.delete_files(definition.workspace, Privacy.HIGH, obsolete)
            if errors:
                log.error(
                    f"Failed to delete high privacy files from workspace {definition.workspace}: {errors}"
                )
            api.delete_files(definition.workspace, Privacy.MEDIUM, obsolete)
            if errors:
                log.error(
                    f"Failed to delete medium privacy files from workspace {definition.workspace}: {errors}"
                )
        mark_job_as_completed(job)
        api.cleanup(definition)
        # we are done here
        return

    # following logic is common to all non-final transitions

    if new_status.state == initial_status.state:
        # no change in state, i.e. back pressure
        set_message(
            job,
            "Waiting on available resources",
            code=StatusCode.WAITING_ON_WORKERS,
        )

    elif new_status.state == expected_state:
        # successful state change to the expected next state
        if new_status.state == ExecutorState.PREPARING:
            job.state = State.RUNNING
            job.started_at = int(time.time())
        elif job.state != State.RUNNING:
            # got an ExecutorState that should mean the job.state is RUNNING, but it is not
            log.warning(
                f"state error: got {new_status.state} for job we thought was {job.state}"
            )
        set_message(job, new_status.state.value.title())

    elif new_status.state == ExecutorState.ERROR:
        # all transitions can go straight to error
        mark_job_as_failed(job, new_status.message)
        api.cleanup(definition)

    else:
        raise InvalidTransition(
            f"unexpected state transition of job {job.id} from {initial_status.state} to {new_status.state}: {new_status.message}"
        )


def save_results(job, results):
    """Extract the results of the execution and update the job accordingly."""
    # set the final state of the job
    if results.exit_code != 0:
        job.state = State.FAILED
        job.status_message = f"Job exited with error code {results.exit_code}"
        job.status_code = StatusCode.NONZERO_EXIT
        if results.message:
            job.status_message += f": {results.message}"
    elif results.unmatched_patterns:
        job.state = State.FAILED
        job.status_message = "No outputs found matching patterns:\n - {}".format(
            "\n - ".join(results.unmatched_patterns)
        )
        # If the job fails because an output was missing its very useful to
        # show the user what files were created as often the issue is just a
        # typo

        # Can we figure these out from job.outputs and project.yaml? Do we do
        # it here or just in local run?
        # TODO:  job.unmatched_outputs = ???
    else:
        job.state = State.SUCCEEDED
        job.status_message = "Completed successfully"

    job.outputs = results.outputs
    job.updated_at = int(time.time())
    update(job)


def get_obsolete_files(definition, outputs):
    """Get files that need to be deleted.

    These are files that we previously output by this action but were not
    output by the latest execution of it, so they've been removed or renamed.

    It does case insenstive comparison, as we don't know the the filesystems
    these will end up being stored on.
    """
    keep_files = {str(name).lower() for name in outputs}
    obsolete = []

    for existing in list_outputs_from_action(definition.workspace, definition.action):
        name = str(existing).lower()
        if name not in keep_files:
            obsolete.append(str(existing))
    return obsolete


def job_to_job_definition(job):

    action_args = shlex.split(job.run_command)
    allow_database_access = False
    env = {"OPENSAFELY_BACKEND": config.BACKEND}
    # Check `is True` so we fail closed if we ever get anything else
    if requires_db_access(action_args) is True:
        if not config.USING_DUMMY_DATA_BACKEND:
            allow_database_access = True
            env["DATABASE_URL"] = config.DATABASE_URLS[job.database_name]
            if config.TEMP_DATABASE_NAME:
                env["TEMP_DATABASE_NAME"] = config.TEMP_DATABASE_NAME
            if config.PRESTO_TLS_KEY and config.PRESTO_TLS_CERT:
                env["PRESTO_TLS_CERT"] = config.PRESTO_TLS_CERT
                env["PRESTO_TLS_KEY"] = config.PRESTO_TLS_KEY
            if config.EMIS_ORGANISATION_HASH:
                env["EMIS_ORGANISATION_HASH"] = config.EMIS_ORGANISATION_HASH
    # Prepend registry name
    image = action_args.pop(0)
    full_image = f"{config.DOCKER_REGISTRY}/{image}"
    if image.startswith("stata-mp"):
        env["STATA_LICENSE"] = str(config.STATA_LICENSE)

    # Jobs which are running reusable actions pull their code from the reusable
    # action repo, all other jobs pull their code from the study repo
    study = Study(job.action_repo_url or job.repo_url, job.action_commit or job.commit)
    # Both of action commit and repo_url should be set if either are
    assert bool(job.action_commit) == bool(job.action_repo_url)

    input_files = []
    for action in job.requires_outputs_from:
        for filename in list_outputs_from_action(job.workspace, action):
            input_files.append(filename)

    outputs = {}
    for privacy_level, named_patterns in job.output_spec.items():
        for name, pattern in named_patterns.items():
            outputs[pattern] = privacy_level

    return JobDefinition(
        job.id,
        study,
        job.workspace,
        job.action,
        full_image,
        action_args,
        env,
        input_files,
        outputs,
        allow_database_access,
    )


def get_states_of_awaited_jobs(job):
    job_ids = job.wait_for_job_ids
    if not job_ids:
        return []

    log.debug("Querying database for state of dependencies")
    states = select_values(Job, "state", id__in=job_ids)
    log.debug("Done query")
    return states


def mark_job_as_failed(job, error, code=None):
    if isinstance(error, str):
        message = error
    else:
        message = f"{type(error).__name__}: {error}"
    if job.cancelled:
        message = "Cancelled by user"
        code = StatusCode.CANCELLED_BY_USER
    set_state(job, State.FAILED, message, code=code)


def mark_job_as_running(job, message="Running"):
    set_state(job, State.RUNNING, message)


def mark_job_as_completed(job):
    # Completed means either SUCCEEDED or FAILED. We just save the job to the
    # database exactly as is with the exception of setting the completed at
    # timestamp
    assert job.state in [State.SUCCEEDED, State.FAILED]
    if job.state == State.FAILED and job.cancelled:
        job.status_message = "Cancelled by user"
        job.status_code = StatusCode.CANCELLED_BY_USER
    job.completed_at = int(time.time())
    log.debug("Updating full job record")
    update_job(job)
    log.debug("Update done")
    log.info(job.status_message, extra={"status_code": job.status_code})


def set_state(job, state, message, code=None):
    timestamp = int(time.time())
    if state == State.RUNNING:
        job.started_at = timestamp
    elif state == State.FAILED or state == State.SUCCEEDED:
        job.completed_at = timestamp
    job.state = state
    job.status_message = message
    job.status_code = code
    job.updated_at = timestamp
    log.debug("Updating job status and timestamps")
    update_job(job)
    log.debug("Update done")
    log.info(job.status_message, extra={"status_code": job.status_code})


def set_message(job, message, code=None):
    timestamp = int(time.time())
    # If message has changed then update and log
    if job.status_message != message:
        job.status_message = message
        job.status_code = code
        job.updated_at = timestamp
        update_job(job)
        log.info(job.status_message, extra={"status_code": job.status_code})
    # If the status message hasn't changed then we only update the timestamp
    # once a minute. This gives the user some confidence that the job is still
    # active without writing to the database every single time we poll
    elif timestamp - job.updated_at >= 60:
        job.updated_at = timestamp
        log.debug("Updating job timestamp")
        update_job(job)
        log.debug("Update done")
        # For long running jobs we don't want to fill the logs up with "Job X
        # is still running" messages, but it is useful to have semi-regular
        # confirmations in the logs that it is still running. The below will
        # log approximately once every 10 minutes.
        if datetime.datetime.fromtimestamp(timestamp).minute % 10 == 0:
            log.info(job.status_message, extra={"status_code": job.status_code})


def get_reason_job_not_started(job):
    log.debug("Querying for running jobs")
    running_jobs = find_where(Job, state=State.RUNNING)
    log.debug("Query done")
    used_resources = sum(
        get_job_resource_weight(running_job) for running_job in running_jobs
    )
    required_resources = get_job_resource_weight(job)
    if used_resources + required_resources > config.MAX_WORKERS:
        if required_resources > 1:
            return "Waiting on available workers for resource intensive job"
        else:
            return "Waiting on available workers"


def get_job_resource_weight(job, weights=config.JOB_RESOURCE_WEIGHTS):
    """
    Get the job's resource weight by checking its workspace and action against
    the config file, default to 1 otherwise
    """
    action_patterns = weights.get(job.workspace)
    if action_patterns:
        for pattern, weight in action_patterns.items():
            if pattern.fullmatch(job.action):
                return weight
    return 1


def update_job(job):
    # The cancelled field is written by the sync thread and we should never update it. The sync thread never updates
    # any other fields after it has created the job, so we're always safe to modify them.
    update(job, exclude_fields=["cancelled"])


if __name__ == "__main__":
    configure_logging()

    try:
        main()
    except KeyboardInterrupt:
        sys.exit(0)
